# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex climate networks.
"""

from typing import Tuple
from collections.abc import Hashable

import numpy as np
from numpy import random

from ..core import Data
from ..core.cache import Cached


class ClimateData(Data, Cached):

    """
    Encapsulates spatio-temporal climate data.

    Provides methods to manipulate this data, i.e. calculate daily (monthly)
    mean values and anomaly values.

    @ivar data_source: (string) - The name of the data source
                                  (model, reanalysis, station)
    """

    #
    #  Defines internal methods
    #

    def __init__(self, observable, grid, time_cycle, anomalies=False,
                 observable_name="", observable_long_name=None, window=None,
                 silence_level=0):
        """
        Initialize an instance of ClimateData.

        The spatio-temporal window is described by the following
        dictionary::

            window = {"time_min": 0., "time_max": 0., "lat_min": 0.,
                      "lat_max": 0., "lon_min": 0., "lon_max": 0.}

        :type observable: 2D array [time, index]
        :arg observable: The array of time series to be represented by the
            :class:`.Data` instance.
        :type grid: :class:`.Grid2D` instance
        :arg grid: The Grid representing the spatial coordinates associated to
            the time series and their temporal sampling.
        :arg int time_cycle: The annual cycle length of the data (units of
            samples).
        :arg bool anomalies: Indicates whether the data are climatological
            anomaly values.
        :arg str observable_name: A short name for the observable.
        :arg str observable_long_name: A long name for the observable.
        :arg dict window: Spatio-temporal window to select a view on the data.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        self._mut_window = 0
        """mutation count"""

        Data.__init__(self, observable=observable, grid=grid,
                      observable_name=observable_name,
                      observable_long_name=observable_long_name,
                      window=window, silence_level=silence_level)

        #  Set class variables
        self.time_cycle = time_cycle
        """(number (int)) - The annual cycle length of the data
                            (units of samples)."""

        self.data_source = ""

        # If data are anomalies skip automatic calculation of anomalies
        self.anomalies = anomalies

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        # The following attributes are assumed immutable:
        #   (_full_observable)
        return (self._mut_window,)

    def __str__(self):
        """
        Returns a string representation.
        """
        return 'ClimateData:\n' + Data.__str__(self)

    #
    #  Define alternative constructors
    #

    @classmethod
    def Load(cls, file_name, observable_name, file_type="NetCDF",
             dimension_names=None, window=None, vertical_level=None,
             silence_level=0, time_cycle=None, data_source=None):
        """
        Initialize an instance of ClimateData.

        Supported file types ``file_type`` are:
          - "NetCDF" for regular (rectangular) grids
          - "iNetCDF" for irregular (e.g. geodesic) grids or station data.

        The :index:`spatio-temporal window` is described by the following
        dictionary::

            window = {"time_min": 0., "time_max": 0., "lat_min": 0.,
                      "lat_max": 0., "lon_min": 0., "lon_max": 0.}

        :arg str file_name: The name of the data file.
        :arg str observable_name: The short name of the observable within data
            file (particularly relevant for NetCDF).
        :arg str file_type: The format of the data file.
        :arg dict dimension_names: The names of the dimensions as used in the
            NetCDF file. Default: {"lat": "lat", "lon": "lon", "time": "time"}
        :arg dict window: Spatio-temporal window to select a view on the data.
        :arg int vertical_level: The vertical level to be extracted from the
            data file. Is ignored for horizontal data sets. If None, the first
            level in the data file is chosen.
        :arg int silence_level: The inverse level of verbosity of the object.
        :arg int time_cycle: The annual cycle length of the data (units of
            samples). NOTE: This is a required argument!
        :arg str data_source: The name of the data source (model, reanalysis,
            station).
        """
        if time_cycle is None:
            raise TypeError("ClimateData.Load() is missing required "
                            "keyword argument: 'time_cycle'")

        if dimension_names is None:
            dimension_names = {"lat": "lat", "lon": "lon", "time": "time"}

        #  Load data using _load_data method from parent class
        res = cls._load_data(file_name=file_name, file_type=file_type,
                             dimension_names=dimension_names,
                             observable_name=observable_name,
                             vertical_level=vertical_level)

        #  Create instance of ClimateData
        data = cls(observable=res["observable"], grid=res["grid"],
                   time_cycle=time_cycle,
                   observable_name=res["observable_name"],
                   observable_long_name=res["observable_long_name"],
                   window=window, silence_level=silence_level)

        #  Set class variables
        data.file_name = file_name
        data.file_type = file_type
        data.vertical_level = vertical_level
        data.data_source = data_source

        return data

    @staticmethod
    def SmallTestData():
        """
        Return test data set of 6 time series with 10 sampling points each.

        **Example:**

        >>> r(Data.SmallTestData().observable())
        array([[ 0.    ,  1.    ,  0.    , -1.    , -0.    ,  1.    ],
               [ 0.309 ,  0.9511, -0.309 , -0.9511,  0.309 ,  0.9511],
               [ 0.5878,  0.809 , -0.5878, -0.809 ,  0.5878,  0.809 ],
               [ 0.809 ,  0.5878, -0.809 , -0.5878,  0.809 ,  0.5878],
               [ 0.9511,  0.309 , -0.9511, -0.309 ,  0.9511,  0.309 ],
               [ 1.    ,  0.    , -1.    , -0.    ,  1.    ,  0.    ],
               [ 0.9511, -0.309 , -0.9511,  0.309 ,  0.9511, -0.309 ],
               [ 0.809 , -0.5878, -0.809 ,  0.5878,  0.809 , -0.5878],
               [ 0.5878, -0.809 , -0.5878,  0.809 ,  0.5878, -0.809 ],
               [ 0.309 , -0.9511, -0.309 ,  0.9511,  0.309 , -0.9511]])

        :rtype: ClimateData instance
        :return: a ClimateData instance for testing purposes.
        """
        data = Data.SmallTestData()

        return ClimateData(observable=data.observable(), grid=data.grid,
                           time_cycle=5, silence_level=2)

    #
    #  Define methods to work with (climatological) anomaly data
    #

    def phase_indices(self):
        """
        Return time indices associated to all phases in the annual cycle.

        In other words, provides all time indices falling into a particular
        day, month etc. of the year.

        Just includes measurements from years for which complete data exists.

        .. note::
           Only the currently selected spatio-temporal window is considered.

        .. note::
           Only the currently selected spatio-temporal window is considered.

        **Example:**

        >>> ClimateData.SmallTestData().phase_indices()
        array([[0, 5], [1, 6], [2, 7], [3, 8], [4, 9]])

        :rtype: 2D Numpy array (int) [phase index, year]
        :return: the time indices associated to all phases of the annual cycle.
        """
        range_years = int(self.grid.grid_size()["time"]
                          / self.time_cycle)

        phase_indices = np.zeros((self.time_cycle, range_years), dtype=int)

        for i in range(self.time_cycle):
            phase_indices[i, :] = np.arange(i, range_years * self.time_cycle,
                                            self.time_cycle)

        return phase_indices

    def indices_selected_phases(self, selected_phases):
        """
        Return sorted time indices associated to certain phase indices.

        .. note::
           Only the currently selected spatio-temporal window is considered.

        **Example:**

        >>> ClimateData.SmallTestData().indices_selected_phases([0,1,4])
        array([0, 1, 4, 5, 6, 9])

        :arg [int] selected_phases: The selected phase indices.
        :rtype: 1D array (int)
        :return: the sorted time indices corresponding to chosen phase indices.
        """
        #  Get all
        phase_indices = self.phase_indices()

        #  Select time indices corresponding to chosen phase indices
        selected_indices = phase_indices[selected_phases, :]

        #  Flatten and sort selected time indices
        selected_indices = selected_indices.flatten()
        selected_indices.sort()

        return selected_indices

    def indices_selected_months(self, selected_months):
        """
        Return sorted time indices associated to certain months.

        Currently, only cycle lengths of 12 (monthly data) and 360
        (standardized daily data) are supported.

        .. note::
           Only the currently selected spatio-temporal window is considered.

        :arg [number] selected_months: The selected months.
        :rtype: 1D array (int)
        :return: the sorted time indices corresponding to chosen months.
        """
        if self.time_cycle == 12:
            return self.indices_selected_phases(selected_months)
        elif self.time_cycle == 360:
            selected_days = []
            for month in selected_months:
                for day in range(30):
                    selected_days.append(month * 30 + day)

            return self.indices_selected_phases(selected_days)
        else:
            raise NotImplementedError("Currently only time cycles 12 and 360 \
                                      are supported")

    @Cached.method(name="climatological mean values")
    def phase_mean(self):
        """
        Calculate mean values of observable for each phase of the annual cycle.

        This is also commonly referred to as climatological mean, e.g., the
        mean temperature for all Januaries in the data set for monthly time
        resolution (time_cycle=12).

        .. note::
           Only the currently selected spatio-temporal window is considered.

        :rtype: 2D Numpy array [cycle index, node index]
        :return: the mean values of observable for each phase of the annual
                 cycle.

        **Example:**

        >>> r(ClimateData.SmallTestData().phase_mean())
        array([[ 0.5   ,  0.5   , -0.5   , -0.5   ,  0.5   ,  0.5   ],
               [ 0.63  ,  0.321 , -0.63  , -0.321 ,  0.63  ,  0.321 ],
               [ 0.6984,  0.1106, -0.6984, -0.1106,  0.6984,  0.1106],
               [ 0.6984, -0.1106, -0.6984,  0.1106,  0.6984, -0.1106],
               [ 0.63  , -0.321 , -0.63  ,  0.321 ,  0.63  , -0.321 ]])
        """
        observable = self.observable()
        time_cycle = self.time_cycle
        N = observable.shape[1]
        phase_mean = np.zeros((time_cycle, N))

        #  Calculate mean value for each day (month) on each node
        for i in range(time_cycle):
            phase_mean[i, :] = observable[i::time_cycle, :].mean(axis=0)
        return phase_mean

    @Cached.method(name="daily (monthly) anomaly values")
    def anomaly(self):
        """
        Calculate anomaly time series from observable.

        To obtain climatological anomaly time series, the climatological means
        are subtracted from each sample in the original time series. This
        procedure is also known as phase averaging.

        .. note::
           Only the currently selected spatio-temporal window is considered.

        :rtype: 2D Numpy array [time, node index]
        :return: the anomalized time series.

        **Example:**

        >>> r(ClimateData.SmallTestData().anomaly()[:,0])
        array([-0.5 , -0.321 , -0.1106,  0.1106,  0.321 ,
                0.5 ,  0.321 ,  0.1106, -0.1106, -0.321 ])
        """
        # If data are anomalies skip automatic calculation of anomalies
        if self.anomalies:
            return self._full_observable

        observable = self.observable()
        time_cycle = self.time_cycle
        anomaly = np.zeros(observable.shape)

        #  Thanks to Jakob Runge
        for i in range(time_cycle):
            sample = observable[i::time_cycle, :]
            anomaly[i::time_cycle, :] = sample - sample.mean(axis=0)
        return anomaly

    def anomaly_selected_months(self, selected_months):
        """
        Return anomaly time series from observable for selected months.

        For further comments, see :meth:`anomaly`.

        .. note::
           Only the currently selected spatio-temporal window is considered.

        :arg [number] selected_months: The selected months.
        :rtype:  2D array [time, node index]
        :return: the anomalized time series for selected months.
        """
        selected_indices = self.indices_selected_months(selected_months)
        print(selected_indices)
        return self.anomaly()[selected_indices, :]

    def shuffled_anomaly(self):
        """
        Return the randomly shuffled anomaly time series.

        Each anomaly time series is shuffled individually.

        .. note::
           Only the currently selected spatio-temporal window is considered.

        **Example** (Anomaly with and without temporal shuffling should have
        the same standard deviation along time axis):

        >>> r(ClimateData.SmallTestData().anomaly().std(axis=0))
        array([ 0.31 , 0.6355, 0.31 , 0.6355, 0.31 , 0.6355])
        >>> r(ClimateData.SmallTestData().shuffled_anomaly().std(axis=0))
        array([ 0.31 , 0.6355, 0.31 , 0.6355, 0.31 , 0.6355])

        :rtype: 2D Numpy array [time, node index]
        :return: the anomalized and shuffled time series.
        """
        if self.silence_level <= 1:
            print("Shuffling anomaly time series for significance tests...")

        N = self.grid.grid_size()["space"]
        shuffled_anomaly = np.empty(self.anomaly().shape)

        for i in range(N):
            temp = self.anomaly()[:, i].copy()
            random.shuffle(temp)
            shuffled_anomaly[:, i] = temp

        return shuffled_anomaly

    def set_window(self, window):
        """
        Set spatio-temporal window.

        Calls set_window method of parent class Data and additionally sets
        flags, so that measures derived from data (mean, anomaly) will be
        recalculated for new window.

        The spatio-temporal window is described by the following dictionary::

           window = {"time_min": 0., "time_max": 0., "lat_min": 0.,
                     "lat_max": 0., "lon_min": 0., "lon_max": 0.}

        If the temporal boundaries are equal, the data's full time range is
        selected. If any of the two corresponding spatial boundaries are
        equal, the data's full spatial extension is included.

        For more information see :meth:`pyunicorn.Data.set_window`.

        **Example:**

        >>> data = ClimateData.SmallTestData()
        >>> data.set_window(window={"time_min": 0., "time_max": 0.,
        ...                 "lat_min": 10., "lat_max": 20.,
        ...                 "lon_min": 5.,  "lon_max": 10.})
        >>> r(data.anomaly())
        array([[ 0.5   , -0.5   ], [ 0.321 , -0.63  ], [ 0.1106, -0.6984],
               [-0.1106, -0.6984], [-0.321 , -0.63  ], [-0.5   ,  0.5   ],
               [-0.321 ,  0.63  ], [-0.1106,  0.6984], [ 0.1106,  0.6984],
               [ 0.321 ,  0.63  ]])

        :type window: dictionary
        :arg window: The spatio-temporal window to select a view on the data.
        """
        Data.set_window(self, window)
        # invalidate cache
        self._mut_window += 1

    def set_global_window(self):
        """
        Set the view on the whole data set.

        Select the full data set and creates a data array as well as
        a corresponding Grid2D object to access this window from outside.

        **Example** (Set smaller window and subsequently restore global
        window):

        >>> data = ClimateData.SmallTestData()
        >>> data.set_window(window={"time_min": 0., "time_max": 4.,
        ...                 "lat_min": 10., "lat_max": 20.,
        ...                 "lon_min": 5.,  "lon_max": 10.})
        >>> data.grid.grid()["lat"]
        array([ 10.,  15.], dtype=float32)
        >>> data.set_global_window()
        >>> data.grid.grid()["lat"]
        array([  0.,   5.,  10.,  15.,  20.,  25.], dtype=float32)
        """
        Data.set_global_window(self)
        # invalidate cache
        self._mut_window += 1



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"
"""
Simple tests for the ResNetwork class.
"""
import numpy as np

from pyunicorn import ResNetwork

# -----------------------------------------------------------------------------
# Class member tests with TestNetwork
# -----------------------------------------------------------------------------


def test_init(capsys):
    print(ResNetwork.SmallTestNetwork())
    out, err = capsys.readouterr()
    out_ref = "ResNetwork:\nGeoNetwork:\nSpatialNetwork:\n" + \
              "Network: undirected, 5 nodes, 5 links, link density 0.500." + \
              "\nGeographical boundaries:\n" + \
              "         time     lat     lon\n" + \
              "   min    0.0    0.00 -180.00\n" + \
              "   max    9.0   90.00  180.00\n" + \
              "Average resistance: 2.4\n"
    assert out == out_ref


def test_SmallTestNetwork():
    assert isinstance(ResNetwork.SmallTestNetwork(), ResNetwork)


def test_SmallComplexNetwork():
    net = ResNetwork.SmallComplexNetwork()
    assert net.flagComplex

    adm = net.get_admittance()

    res = adm.real
    exp = [[0., 0.1, 0., 0., 0.],
           [0.1, 0., 0.0625, 0.25, 0.],
           [0., 0.0625, 0., 0.0625, 0.],
           [0., 0.25, 0.0625, 0., 0.05],
           [0., 0., 0., 0.05, 0.]]
    assert np.allclose(res, exp, atol=1e-04)

    res = adm.imag
    exp = [[0., -0.2, 0., 0., 0.],
           [-0.2, 0., -0.0625, -0.25, 0.],
           [0., -0.0625, 0., -0.0625, 0.],
           [0., -0.25, -0.0625, 0., -0.05],
           [0., 0., 0., -0.05, 0.]]
    assert np.allclose(res, exp, atol=1e-04)


def test_update_resistances():
    net = ResNetwork.SmallTestNetwork()
    net.update_resistances(net.adjacency)

    res = net.get_admittance()
    exp = [[0., 1., 0., 0., 0.],
           [1., 0., 1., 1., 0.],
           [0., 1., 0., 1., 0.],
           [0., 1., 1., 0., 1.],
           [0., 0., 0., 1., 0.]]
    assert (res == exp).all()

    res = net.admittance_lapacian()
    exp = [[1., -1., 0., 0., 0.],
           [-1., 3., -1., -1., 0.],
           [0., -1., 2., -1., 0.],
           [0., -1., -1., 3., -1.],
           [0., 0., 0., -1., 1.]]
    assert (res == exp).all()


def test_update_admittance():
    net = ResNetwork.SmallTestNetwork()
    net.update_admittance()
    assert True


def test_get_admittance():
    net = ResNetwork.SmallTestNetwork()

    res = net.get_admittance()
    exp = [[0., 0.5, 0., 0., 0.],
           [0.5, 0., 0.125, 0.5, 0.],
           [0., 0.125, 0., 0.125, 0.],
           [0., 0.5, 0.125, 0., 0.1],
           [0., 0., 0., 0.1, 0.]]
    assert np.allclose(res, exp, atol=1e-04)


def test_update_R():
    net = ResNetwork.SmallTestNetwork()
    net.update_R()
    assert True


def test_get_R():
    res = ResNetwork.SmallTestNetwork().get_R()
    exp = [[2.28444444, 0.68444444, -0.56, -0.20444444, -2.20444444],
           [0.68444444, 1.08444444, -0.16, 0.19555556, -1.80444444],
           [-0.56, -0.16, 3.04, -0.16, -2.16],
           [-0.20444444, 0.19555556, -0.16, 1.08444444, -0.91555556],
           [-2.20444444, -1.80444444, -2.16, -0.91555556, 7.08444444]]
    assert np.allclose(res, exp, atol=1e-04)


def test_admittance_laplacian():
    res = ResNetwork.SmallTestNetwork().admittance_lapacian()
    exp = [[0.5, -0.5, 0., 0., 0.],
           [-0.5, 1.125, -0.125, -0.5, 0.],
           [0., -0.125, 0.25, -0.125, 0.],
           [0., -0.5, -0.125, 0.725, -0.1],
           [0., 0., 0., -0.1, 0.1]]
    assert np.allclose(res, exp, atol=1e-04)


def test_admittive_degree():
    res = ResNetwork.SmallTestNetwork().admittive_degree()
    exp = [0.5, 1.125, 0.25, 0.725, 0.1]
    assert np.allclose(res, exp, atol=1e-04)


def test_average_neighbors_admittive_degree():
    res = ResNetwork.SmallTestNetwork().average_neighbors_admittive_degree()
    exp = [2.25, 1.31111111, 7.4, 2.03448276, 7.25]
    assert np.allclose(res, exp, atol=1e-04)


def test_local_admittive_clustering():
    res = ResNetwork.SmallTestNetwork().local_admittive_clustering()
    exp = [0., 0.00694444, 0.0625, 0.01077586, 0.]
    assert np.allclose(res, exp, atol=1e-04)


def test_global_admittive_clustering():
    res = ResNetwork.SmallTestNetwork().global_admittive_clustering()
    exp = 0.016
    assert np.isclose(res, exp, atol=1e-04)


def test_effective_resistance():
    net = ResNetwork.SmallTestNetwork()

    res = net.effective_resistance(1, 1)
    exp = 0.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.effective_resistance(1, 2)
    exp = 4.4444
    assert np.isclose(res, exp, atol=1e-04)


def test_average_effective_resistance():
    res = ResNetwork.SmallTestNetwork().average_effective_resistance()
    exp = 7.2889
    assert np.isclose(res, exp, atol=1e-04)


def test_diameter_effective_resistance():
    res = ResNetwork.SmallTestNetwork().diameter_effective_resistance()
    exp = 14.4444
    assert np.isclose(res, exp, atol=1e-04)


def test_effective_resistance_closeness_centrality():
    net = ResNetwork.SmallTestNetwork()

    res = net.effective_resistance_closeness_centrality(0)
    exp = 0.1538
    assert np.isclose(res, exp, atol=1e-04)

    res = net.effective_resistance_closeness_centrality(4)
    exp = 0.08
    assert np.isclose(res, exp, atol=1e-04)


def test_vertex_current_flow_betweenness():
    net = ResNetwork.SmallTestNetwork()

    res = net.vertex_current_flow_betweenness(1)
    exp = 0.3889
    assert np.isclose(res, exp, atol=1e-04)


def test_edge_current_flow_betweenness():
    net = ResNetwork.SmallTestNetwork()

    res = net.edge_current_flow_betweenness()
    exp = [[0., 0.4, 0., 0., 0.],
           [0.4, 0., 0.2444, 0.5333, 0.],
           [0., 0.2444, 0., 0.2444, 0.],
           [0., 0.5333, 0.2444, 0., 0.4],
           [0., 0., 0., 0.4, 0.]]
    assert np.allclose(res, exp, atol=1e-04)

    net.update_resistances(net.adjacency)
    res = net.edge_current_flow_betweenness()
    exp = [[0., 0.4, 0., 0., 0.],
           [0.4, 0., 0.3333, 0.4, 0.],
           [0., 0.3333, 0., 0.3333, 0.],
           [0., 0.4, 0.3333, 0., 0.4],
           [0., 0., 0., 0.4, 0.]]
    assert np.allclose(res, exp, atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Tests for the EventSeries class.
"""
import numpy as np
from pyunicorn.eventseries import EventSeries


def create_test_data():
    N = 1000
    tstamps = np.linspace(0.0, (N-1)*0.1, N)
    data = np.zeros((N, 2))
    for i in range(N):
        data[i] = np.array([np.sin(tstamps[i]*np.pi),
                            np.sin((tstamps[i]-0.2)*np.pi)])
    return data, tstamps


def test_EventSeries_init():
    data = np.random.randint(2, size=(1000, 3))
    test_object1 = EventSeries(data)

    assert np.all(test_object1.get_event_matrix() == data)


def test_make_event_matrix():
    data, tstamps = create_test_data()

    # Test 'value' method for first variable > 0.99 and second variable < 0.0
    thresholds1 = np.array([0.99, 0.0])
    test_object1 = EventSeries(data, threshold_method='value',
                               threshold_values=thresholds1,
                               threshold_types=np.array(['above', 'below']))
    eventmatrix = np.zeros((data.shape[0], data.shape[1]))
    for i in range(data.shape[0]):
        if data[i][0] > thresholds1[0]:
            eventmatrix[i][0] = 1
        else:
            eventmatrix[i][0] = 0

        if data[i][1] < thresholds1[1]:
            eventmatrix[i][1] = 1
        else:
            eventmatrix[i][1] = 0

    assert np.all(test_object1.get_event_matrix() == eventmatrix)

    # Test 'quantile' method for first variable < 0.3 quantile and
    # second variable > 0.8 quantile
    threshold_values2 = np.array([0.3, 0.8])
    test_object2 = EventSeries(data, threshold_method='quantile',
                               threshold_values=threshold_values2,
                               threshold_types=['below', 'above'])
    eventmatrix = np.zeros((data.shape[0], data.shape[1]))
    thresholds2 = [np.quantile(data[:, 0], threshold_values2[0]),
                   np.quantile(data[:, 1], threshold_values2[1])]
    for i in range(data.shape[0]):
        if data[i][0] < thresholds2[0]:
            eventmatrix[i][0] = 1
        else:
            eventmatrix[i][0] = 0

        if data[i][1] > thresholds2[1]:
            eventmatrix[i][1] = 1
        else:
            eventmatrix[i][1] = 0

    assert np.all(test_object2.get_event_matrix() == eventmatrix)


def eca_second_implementaion(eventseriesx, eventseriesy, ts1=None,
                             ts2=None, deltaT=3, lag=0.0):
    """
    Test implementation of event coincidence analysis

    :type eventseriesx: 1D numpy array
    :arg eventseriesx: Event series containing '0' and '1'
    :type eventseriesy: 1D numpy array
    :arg eventseriesy: Event series containing '0' and '1'
    :type ts1: 1D numpy array
    :arg ts1: time stamps of events in eventseriesx, not obligatory
    :type ts2: 1D numpy array
    :arg ts2: time stamps of events in eventseriesy, not obligatory
    :type deltaT: float
    :arg deltaT: time window defining if two events are considered precursor
                 or trigger of the other
    :type lag: float
    :arg lag: delay between the two time series, eventseriesy is shifted
    :rtype: list
    :return:
    """

    # Get time indices
    if ts1 is None:
        e1 = np.where(eventseriesx)[0]
    else:
        e1 = ts1[eventseriesx == 1]
    if ts2 is None:
        e2 = np.where(eventseriesy)[0]
    else:
        e2 = ts2[eventseriesy == 1]

    # Number of events
    l1 = len(e1)
    l2 = len(e2)

    # Find number of events which cannot be trigger or precursor events
    s1p = len(e1[e1 <= (e1[0] + lag + deltaT)])
    s2t = len(e2[e2 >= (e2[-1] - lag - deltaT)])
    s1t = len(e1[e1 >= (e1[-1] - lag - deltaT)])
    s2p = len(e2[e2 <= (e2[0] + lag + deltaT)])

    precursorxy = 0.0
    precursoryx = 0.0
    triggerxy = 0.0
    triggeryx = 0.0

    for i1 in range(s1p, l1):
        precursordummy = False
        for j2 in range(l2):
            if e1[i1] - lag - e2[j2] >= 0.0 and \
                    e1[i1] - lag - e2[j2] <= deltaT:
                precursordummy = True
        if precursordummy:
            precursorxy = precursorxy + 1.0

    for j2 in range(l2-s2t):
        triggerdummy = False
        for i1 in range(l1):
            if e1[i1] - lag - e2[j2] >= 0.0 and \
                    e1[i1] - lag - e2[j2] <= deltaT:
                triggerdummy = True
        if triggerdummy:
            triggerxy = triggerxy + 1.0

    for i2 in range(s2p, l2):
        precursordummy = False
        for j1 in range(l1):
            if e2[i2] - lag - e1[j1] >= 0.0 and \
                    e2[i2] - lag - e1[j1] <= deltaT:
                precursordummy = True
        if precursordummy:
            precursoryx = precursoryx + 1.0

    for j1 in range(l1-s1t):
        triggerdummy = False
        for i2 in range(l2):
            if e2[i2] - lag - e1[j1] >= 0.0 and \
                    e2[i2] - lag - e1[j1] <= deltaT:
                triggerdummy = True
        if triggerdummy:
            triggeryx = triggeryx + 1.0

    return (precursorxy/(l1-s1p), triggerxy/(l2-s2t),
            precursoryx/(l2-s2p), triggeryx/(l1-s1t))


def test_eca():
    es1 = np.random.randint(2, size=100)
    es2 = np.random.randint(2, size=100)
    for tau in [0.0, 1.5]:
        for window in [2, 5]:
            for ts1, ts2 in [[None, None],
                             [np.random.uniform(size=len(es1)).cumsum(),
                              np.random.uniform(size=len(es1)).cumsum()]]:
                eca_orig = \
                    EventSeries.event_coincidence_analysis(es1, es2, window,
                                                           ts1=ts1,
                                                           ts2=ts2,
                                                           lag=tau)
                eca_check = \
                    eca_second_implementaion(es1, es2, ts1=ts1, ts2=ts2,
                                             deltaT=window, lag=tau)
                assert np.allclose(np.array(eca_orig), np.array(eca_check),
                                   atol=1e-04)


def eca_implementation_for_symmetric_window(es1, es2, ts1=None, ts2=None,
                                            taumax=3.0, lag=0.0):
    # Get time indices
    if ts1 is None:
        e1 = np.where(es1)[0]
    else:
        e1 = ts1[es1 == 1]
    if ts2 is None:
        e2 = np.where(es2)[0]
    else:
        e2 = ts2[es2 == 1]

    # Number of events
    l1 = len(e1)
    l2 = len(e2)

    # Find number of events which cannot be trigger or precursor events
    if not (lag == 0 and taumax == 0):
        s1start = len(e1[e1 <= (e1[0] + lag + taumax)])
        s1end = len(e1[e1 >= (e1[-1] - lag - taumax)])
        s2end = len(e2[e2 >= (e2[-1] - lag - taumax)])
        s2start = len(e2[e2 <= (e2[0] + lag + taumax)])
    else:
        s1start, s1end, s2start, s2end = 0, 0, 0, 0

    coincidence12, coincidence21 = 0.0, 0.0

    for i in range(s1start, l1 - s1end):
        for j in range(l2):
            if np.abs(e1[i] - lag - e2[j]) <= taumax:
                coincidence12 = coincidence12 + 1.0
                break

    for j in range(s2start, l2 - s2end):
        for i in range(l1):
            if np.abs(e2[j] - lag - e1[i]) <= taumax:
                coincidence21 = coincidence21 + 1.0
                break

    return (coincidence12 / (l1 - s1start - s1end),
            coincidence21 / (l2 - s2start - s2end))


def es_second_implementation(es1, es2, ts1=None, ts2=None, taumax=np.inf,
                             lag=0.0):
    """
    Calculates the directed event synchronization from two event series X
    and Y.
    :type es1: 1D Numpy array
    :arg es1: Event series containing '0's and '1's
    :type es2: 1D Numpy array
    :arg es2: Event series containing '0's and '1's
    :type ts1: 1D Numpy array
    :arg ts1: Event time array containing time points when events of event
               series 1 occur, not obligatory
    :type ts2: 1D Numpy array
    :arg ts2: Event time array containing time points when events of event
               series 2 occur, not obligatory
    :type taumax: float
    :arg taumax: maximum distance of two events to be counted as
                 synchronous
    :type lag: float
    :arg lag: delay between the two event series, the second event series
                  is shifted by the value of lag
    :rtype: list
    :return: [Event synchronization XY, Event synchronization YX]
    """

    # Get time indices
    if ts1 is None:
        ex = np.where(es1)[0]
    else:
        ex = ts1[es1 == 1]
    if ts2 is None:
        ey = np.where(es2)[0]
    else:
        ey = ts2[es2 == 1]

    lx = len(ex)
    ly = len(ey)

    ey = ey + lag

    Axy = np.zeros((lx, ly), dtype=bool)
    Ayx = np.zeros((lx, ly), dtype=bool)
    eqtime = np.zeros((lx, ly), dtype=bool)
    for m in range(1, lx - 1):
        for n in range(1, ly - 1):
            dstxy = ex[m] - ey[n]

            if abs(dstxy) > taumax:
                continue
            # finding the dynamical delay lag
            tau = min([ex[m + 1] - ex[m], ex[m] - ex[m - 1],
                       ey[n + 1] - ey[n], ey[n] - ey[n - 1]]) / 2

            if dstxy > 0:
                if dstxy <= tau:
                    Axy[m, n] = True
            if dstxy < 0:
                if dstxy >= -tau:
                    Ayx[m, n] = True
            elif dstxy == 0:
                eqtime[m, n] = True

    # Loop over coincidences and determine number of double counts
    # by checking at least one event of the pair is also coincided
    # in other direction
    countxydouble = countyxdouble = 0

    for i, j in np.transpose(np.where(Axy)):
        countxydouble += np.any(Ayx[i, :]) or np.any(Ayx[:, j])
    for i, j in np.transpose(np.where(Ayx)):
        countyxdouble += np.any(Axy[i, :]) or np.any(Axy[:, j])

    countxy = np.sum(Axy) + 0.5 * np.sum(eqtime) - 0.5 * countxydouble
    countyx = np.sum(Ayx) + 0.5 * np.sum(eqtime) - 0.5 * countyxdouble
    norm = np.sqrt((lx - 2) * (ly - 2))
    return countxy / norm, countyx / norm


def test_es():
    es1 = np.random.binomial(1, 0.5, 100)
    es2 = np.random.binomial(1, 0.5, 100)

    for tau in [0.0, 1.5]:
        for taumax in [np.inf, 5]:
            for ts1, ts2 in [[None, None],
                             [np.random.uniform(size=len(es1)).cumsum(),
                              np.random.uniform(size=len(es1)).cumsum()]]:
                es_orig = \
                    EventSeries.event_synchronization(es1, es2, ts1=ts1,
                                                      ts2=ts2, taumax=taumax,
                                                      lag=tau)
                es_check = \
                    es_second_implementation(es1, es2, ts1=ts1, ts2=ts2,
                                             taumax=taumax, lag=tau)

                assert np.allclose(np.array(es_orig), np.array(es_check),
                                   atol=1e-04)


def eca_matrix_second_implementation(eventmatrix, taumax,
                                     eca_type):

    if eca_type in ['advanced', 'retarded']:
        N = eventmatrix.shape[1]
        res_p, res_t = np.zeros((N, N)), np.zeros((N, N))

        for i in np.arange(0, N):
            for j in np.arange(i + 1, N):
                res_p[i, j], res_t[i, j], res_p[j, i], res_t[j, i] = \
                    eca_second_implementaion(eventmatrix[:, i],
                                             eventmatrix[:, j],
                                             deltaT=taumax)

        if eca_type == 'advanced':
            return res_p

        elif eca_type == 'retarded':
            return res_t

        else:
            return None

    elif eca_type == 'symmetric':
        N = eventmatrix.shape[1]
        res_s = np.zeros((N, N))
        for i in np.arange(0, N):
            for j in np.arange(i + 1, N):
                res_s[i, j], res_s[j, i] = \
                    eca_implementation_for_symmetric_window(eventmatrix[:, i],
                                                            eventmatrix[:, j],
                                                            taumax=taumax)
        return res_s

    return None


def test_vectorized_eca():
    # Create an event matrix of length 1000 with 3 variables
    test_event_matrix = np.random.randint(2, size=(100, 3))

    for eca_type in ['advanced', 'retarded', 'symmetric']:
        for taumax in [1, 5, 16]:
            eso = EventSeries(test_event_matrix, taumax=taumax)
            directedeca_matrix = \
                eca_matrix_second_implementation(test_event_matrix,
                                                 taumax, eca_type)
            # directed, no symmetry
            assert \
                np.allclose(directedeca_matrix,
                            eso.event_series_analysis(
                                method='ECA', symmetrization='directed',
                                window_type=eca_type),
                            atol=1e-04)
            # Test for the various symmetry options
            # mean
            assert \
                np.allclose(np.mean([directedeca_matrix, directedeca_matrix.T],
                                    axis=0),
                            eso.event_series_analysis(method='ECA',
                                                      symmetrization='mean',
                                                      window_type=eca_type),
                            atol=1e-04)
            # max
            assert \
                np.allclose(np.max([directedeca_matrix, directedeca_matrix.T],
                                   axis=0),
                            eso.event_series_analysis(method='ECA',
                                                      symmetrization='max',
                                                      window_type=eca_type),
                            atol=1e-04)

            # min
            assert \
                np.allclose(np.min([directedeca_matrix, directedeca_matrix.T],
                                   axis=0),
                            eso.event_series_analysis(method='ECA',
                                                      symmetrization='min',
                                                      window_type=eca_type),
                            atol=1e-04, equal_nan=True)


def es_matrix_second_implementation(eventmatrix, taumax):
    N = eventmatrix.shape[1]
    res = np.zeros((N, N))

    for i in np.arange(0, N):
        for j in np.arange(i + 1, N):
            res[i, j], res[j, i] = \
                es_second_implementation(eventmatrix[:, i],
                                         eventmatrix[:, j], taumax=taumax)
    return res


def test_vectorized_es():
    # Test if the vectorized implementation coincides with the straight forward
    # one.

    for taumax in [1, 5, 16, np.inf]:
        length, N, eventprop = 1000, 2, 0.2
        # equal event counts (normalization requirement)
        eventcount = int(length * eventprop)
        eventmatrix = np.zeros((length, N), dtype=int)
        for v in range(N):
            fills = np.random.choice(np.arange(length), eventcount,
                                     replace=False)
            eventmatrix[fills, v] = 1

        esob = EventSeries(eventmatrix, taumax=taumax)
        directedes_matrix = es_matrix_second_implementation(eventmatrix,
                                                            taumax)
        # directed, no symmetry
        assert \
            np.allclose(directedes_matrix,
                        esob.event_series_analysis(method='ES',
                                                   symmetrization='directed'),
                        atol=1e-04)

        # Test for all symmetry options
        # symmetric
        assert \
            np.allclose(directedes_matrix + directedes_matrix.T,
                        esob.event_series_analysis(method='ES',
                                                   symmetrization='symmetric'),
                        atol=1e-04)

        # antisymmetric
        assert \
            np.allclose(directedes_matrix - directedes_matrix.T,
                        esob.event_series_analysis(method='ES',
                                                   symmetrization='antisym'),
                        atol=1e-04, equal_nan=True)

        # mean
        assert \
            np.allclose(np.mean([directedes_matrix, directedes_matrix.T],
                                axis=0),
                        esob.event_series_analysis(method='ES',
                                                   symmetrization='mean'),
                        atol=1e-04)

        # max
        assert \
            np.allclose(np.maximum(directedes_matrix, directedes_matrix.T),
                        esob.event_series_analysis(method='ES',
                                                   symmetrization='max'),
                        atol=1e-04)

        # min
        assert \
            np.allclose(np.minimum(directedes_matrix, directedes_matrix.T),
                        esob.event_series_analysis(method='ES',
                                                   symmetrization='min'),
                        atol=1e-04)


def test_significance():
    n_surrs = 100
    data, tstamps = create_test_data()

    # Testing without symmetry
    esob = EventSeries(data, timestamps=tstamps, threshold_method='value',
                       threshold_values=np.array([0.98, 0.98]),
                       threshold_types=np.array(['above', 'above']),
                       taumax=np.inf, lag=0.0)

    # Test Event Synchronization significance via surrogates
    assert \
        np.allclose(esob.event_analysis_significance(method='ES',
                                                     n_surr=n_surrs),
                    np.array([[0., 0.], [1., 0.]]), atol=1e-04)
    # Test ECA
    esob = EventSeries(data, timestamps=tstamps, threshold_method='value',
                       threshold_values=np.array([0.98, 0.98]),
                       threshold_types=np.array(['above', 'above']),
                       taumax=1.0, lag=0.0)

    # Test precursor coincidence via surrogates
    assert \
        np.allclose(esob.event_analysis_significance(method='ECA',
                                                     surrogate='shuffle',
                                                     n_surr=n_surrs,
                                                     symmetrization='directed',
                                                     window_type='advanced'),
                    np.array([[0.0, 0.0], [1.0, 0.0]]), atol=1e-04)

    # Test trigger coincidence via surrogates
    assert \
        np.allclose(esob.event_analysis_significance(method='ECA',
                                                     surrogate='shuffle',
                                                     n_surr=n_surrs,
                                                     symmetrization='directed',
                                                     window_type='retarded'),
                    np.array([[0.0, 0.0], [1.0, 0.0]]), atol=1e-04)
    # Test trigger coincidence via binomial distribution
    assert \
        np.allclose(esob.event_analysis_significance(method='ECA',
                                                     surrogate='analytic',
                                                     symmetrization='directed',
                                                     window_type='retarded'),
                    np.array([[0.0, 0.0], [1.0, 0.0]]), atol=1e-04)

    # Test precursor coincidence via binomial distribution
    assert \
        np.allclose(esob.event_analysis_significance(method='ECA',
                                                     surrogate='analytic',
                                                     symmetrization='directed',
                                                     window_type='advanced'),
                    np.array([[0.0, 0.0], [1.0, 0.0]]), atol=1e-04)

    # Testing for symmetrization=mean
    # Testing without symmetry
    esob = EventSeries(data, timestamps=tstamps, threshold_method='value',
                       threshold_values=np.array([0.98, 0.98]),
                       threshold_types=np.array(['above', 'above']),
                       taumax=np.inf, lag=0.0)

    # Test Event Synchronization significance via surrogates
    assert \
        np.allclose(esob.event_analysis_significance(method='ES',
                                                     n_surr=n_surrs,
                                                     symmetrization='mean'),
                    np.array([[0., 1.0], [1.0, 0.]]), atol=1e-04)
    # Test ECA
    esob = EventSeries(data, timestamps=tstamps, threshold_method='value',
                       threshold_values=np.array([0.98, 0.98]),
                       threshold_types=np.array(['above', 'above']),
                       taumax=1.0, lag=0.0)

    # Test precursor coincidence via surrogates
    assert \
        np.allclose(esob.event_analysis_significance(method='ECA',
                                                     surrogate='shuffle',
                                                     n_surr=n_surrs,
                                                     symmetrization='mean',
                                                     window_type='advanced'),
                    np.array([[0.0, 1.0], [1.0, 0.0]]), atol=1e-04)

    # Test trigger coincidence via surrogates
    assert \
        np.allclose(esob.event_analysis_significance(method='ECA',
                                                     surrogate='shuffle',
                                                     n_surr=n_surrs,
                                                     symmetrization='mean',
                                                     window_type='retarded'),
                    np.array([[0.0, 1.0], [1.0, 0.0]]), atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for analyzing spatially embedded complex networks, handling
multivariate data.
Written by Jakob Runge.
"""

import numpy                        # array object and fast numerics
from scipy import special, linalg   # special math functions

# import mpi                          # parallelized computations

from ..core._ext.types import to_cy, LAG, FIELD, \
    INT16TYPE, INT32TYPE, INT64TYPE
from ._ext.numerics import _symmetrize_by_absmax, _cross_correlation_max, \
    _cross_correlation_all, _get_nearest_neighbors


#
#  Define class Coupling Analysis
#

class CouplingAnalysis:

    """
    Contains methods to calculate coupling matrices from large arrays
    of scalar time series.
    Comprises linear and information-theoretic measures, lagged
    and directed couplings.
    """

    #
    #  Definitions of internal methods
    #

    def __init__(self, data, silence_level=0):
        """
        Initialize an instance of CouplingAnalysis from data array.

        :type data: multidimensional numpy array
        :arg data: The time series array with time in first dimension.
        :type silence_level: int >= 0
        :arg  silence_level: The higher, the less progress info is output.
        """

        self.silence_level = silence_level
        """(int>=0) higher -> less progress info"""

        #  Flatten array along spatial dimensions to allow
        #  for more convinient indexing
        self.n_time = data.shape[0]
        self.data = data.reshape(self.n_time, -1)
        self.N = self.data.shape[1]

        #  precalculation of p*log(p) needed for entropies
        self.plogp = None

    def __str__(self):
        """Return a string representation of the CouplingAnalysis object."""
        return (f'CouplingAnalysis: {self.N} variables, '
                f'{self.n_time} timesteps.')

    @staticmethod
    def test_data():
        """
        Return example test data as discussed in pyunicorn description paper.
        """
        numpy.random.seed(42)
        noise = numpy.random.randn(1000, 4)
        data = noise
        for t in range(2, 1000):
            data[t, 0] = 0.8 * data[t-1, 0] + noise[t, 0]
            data[t, 1] = 0.8 * data[t-1, 1] + 0.5 * data[t-2, 0] + noise[t, 1]
            data[t, 2] = 0.7 * data[t-1, 0] + noise[t, 2]
            data[t, 3] = 0.7 * data[t-2, 0] + noise[t, 3]
        return data

    def symmetrize_by_absmax(self, similarity_matrix, lag_matrix):

        """
        Returns symmetrized similarity matrix.

        Computes the largest absolute value for each pair (i,j) and (j,i) and
        returns the in-place changed matrices of measures and lags. A negative
        lag for an entry (i,j) in the lag_matrix then indicates a 'direction'
        j --> i regarding the peak of the lag function, and vice versa for a
        positive lag.

        **Example:**

        >>> coup_ana = CouplingAnalysis(CouplingAnalysis.test_data())
        >>> similarity_matrix, lag_matrix = coup_ana.cross_correlation(
        ...     tau_max=2)
        >>> r((similarity_matrix, lag_matrix))
        (array([[ 1.    , 0.698 , 0.7788, 0.7535],
                [ 0.4848, 1.    , 0.4507, 0.52  ],
                [ 0.6219, 0.5704, 1.    , 0.5996],
                [ 0.4833, 0.5503, 0.5002, 1.    ]]),
         array([[0, 2, 1, 2], [0, 0, 0, 0],
                [0, 2, 0, 1], [0, 2, 0, 0]]))
        >>> r(coup_ana.symmetrize_by_absmax(similarity_matrix, lag_matrix))
        (array([[ 1.    , 0.698 , 0.7788, 0.7535],
                [ 0.698 , 1.    , 0.5704, 0.5503],
                [ 0.7788, 0.5704, 1.    , 0.5996],
                [ 0.7535, 0.5503, 0.5996, 1.    ]]),
         array([[ 0, 2, 1, 2], [-2, 0, -2, -2],
                [-1, 2, 0, 1], [-2, 2, -1, 0]]))

        :type similarity_matrix: array-like [float]
        :arg  similarity_matrix: array-like [node, node] matrix of similarity
                                 estimates

        :type lag_matrix: array-like [int>=0]
        :arg  lag_matrix:  array-like [node, node] matrix of lags

        :rtype: tuple of arrays
        :returns: the value at the absolute maximum and the (pos or neg) lag.
        """

        return _symmetrize_by_absmax(to_cy(similarity_matrix, FIELD),
                                     to_cy(lag_matrix, LAG), self.N)

    #
    #  Define methods to estimate similarity measures
    #
    def cross_correlation(self, tau_max=0, lag_mode='max'):
        r"""
        Return cross correlation between all pairs of nodes.

        Two lag-modes are available (default: lag_mode='max'):

        lag_mode = 'all':
        Return 3-dimensional array of lagged cross correlations between all
        pairs of nodes. An entry :math:`(i, j, \tau)` corresponds to
        :math:`\rho(X^i_t-\tau, X^j_t)` for positive lags tau, i.e., the
        direction i --> j for :math:`\tau \ne 0`.

        lag_mode = 'max':
        Return matrix of absolute maxima and corresponding lags of lagged
        cross correlation (CC) between all pairs of nodes.
        Returns two usually asymmetric matrices of CC values and lags: In each
        matrix, an entry :math:`(i, j)` corresponds to the (positive or
        negative) value and lag, respectively, at absolute maximum of
        :math:`\rho(X^i_t-\tau, X^j_t)` for positive lags tau, i.e., the
        direction i --> j for :math:`\tau > 0`. The matrices are, thus,
        asymmetric. The function :meth:`.symmetrize_by_absmax` can be used to
        obtain a symmetric matrix.

        **Example:**

        >>> coup_ana = CouplingAnalysis(CouplingAnalysis.test_data())
        >>> similarity_matrix, lag_matrix = coup_ana.cross_correlation(
        ...     tau_max=5, lag_mode='max')
        >>> r((similarity_matrix, lag_matrix))
        (array([[ 1.   ,  0.757 ,  0.779 ,  0.7536],
               [ 0.4847,  1.    ,  0.4502,  0.5197],
               [ 0.6219,  0.5844,  1.    ,  0.5992],
               [ 0.4827,  0.5509,  0.4996,  1.    ]]),
         array([[0, 4, 1, 2], [0, 0, 0, 0], [0, 3, 0, 1], [0, 2, 0, 0]]))

        :type tau_max: int [int>=0]
        :arg  tau_max: maximum lag of cross correlation lag function.

        :type lag_mode: str [('max'|'all')]
        :arg  lag_mode: lag-mode of cross correlations to return.

        :rtype: 3D-array or tuple of matrices
        :returns: all-lag array or matrices of value and lag at the absolute
                  maximum.
        """

        data = self.data
        T, N = data.shape

        # Sanity checks
        if not isinstance(data, numpy.ndarray):
            raise TypeError(f"data is of type {type(data)}, "
                            "must be numpy.ndarray")
        if N > T:
            print(f"Warning: data.shape = {data.shape},"
                  " is it of shape (observations, variables) ?")
        assert numpy.isnan(data).sum() == 0, "NaNs in the data"
        assert tau_max >= 0, f"{tau_max =}"
        assert lag_mode in ['max', 'all'], f"{lag_mode =}"

        #  Normalize time series to zero mean and unit variance for all lags
        corr_range = T - tau_max
        array = numpy.empty((tau_max + 1, N, corr_range), dtype=FIELD)

        for t in range(tau_max + 1):
            #  Remove mean value from time series at each node
            array[t] = (
                data[t:t+corr_range, :]
                - data[t:t+corr_range, :].mean(axis=0).reshape(1, N)).T

            #  Normalize the variance of anomalies to one
            array[t] /= array[t].std(axis=1).reshape(N, 1)

            #  Correct for nodes with zero variance in their time series
            array[t][numpy.isnan(array[t])] = 0

        if lag_mode == 'max':
            return _cross_correlation_max(
                to_cy(array, FIELD), N, tau_max, corr_range)
        elif lag_mode == 'all':
            return _cross_correlation_all(
                to_cy(array, FIELD), N, tau_max, corr_range)
        else:
            return None

    def mutual_information(self, tau_max=0, estimator='knn',
                           knn=10, bins=6, lag_mode='max'):
        r"""
        Return mutual information (MI) between all pairs of nodes.

        Three estimators are available:

        estimator = 'knn' (Recommended):
        Based on k-nearest-neighbors [Kraskov2004]_,
        version 1 in their paper. Larger k have smaller variance, but larger
        (typically negative) bias, and vice versa.

        estimator = 'binning':
        Binning estimator based on equal-quantile binning.

        estimator = 'gauss':
        Captures only linear part of association. Essentially estimates a
        transformed partial correlation.


        Two lag-modes are available (default: lag_mode='max'):

        lag_mode = 'all':
        Return 3-dimensional array of lagged MI between all pairs of nodes. An
        entry :math:`(i, j, \tau)` corresponds to :math:`I(X^i_t-\tau, X^j_t)`
        for positive lags tau, i.e., the direction i --> j for :math:`\tau \ne
        0`.

        lag_mode = 'max':
        Return matrix of absolute maxima and corresponding lags of lagged
        MI between all pairs of nodes.
        Returns two usually asymmetric matrices of MI values and lags: In each
        matrix, an entry :math:`(i, j)` corresponds to the value and lag,
        respectively, at absolute maximum of :math:`I(X^i_t-\tau, X^j_t)` for
        positive lags tau, i.e., the direction i --> j for :math:`\tau > 0`.
        The matrices are, thus, asymmetric. The function
        :meth:`.symmetrize_by_absmax` can be used to obtain a symmetric matrix.

        Reference: [Kraskov2004]_

        **Example:**

        >>> coup_ana = CouplingAnalysis(CouplingAnalysis.test_data())
        >>> similarity_matrix, lag_matrix = coup_ana.mutual_information(
        ...     tau_max=5, knn=10, estimator='knn')
        >>> r(similarity_matrix)
        array([[ 4.6505,  0.4387,  0.4652,  0.4126],
               [ 0.147 ,  4.6505,  0.1065,  0.1639],
               [ 0.2483,  0.2126,  4.6505,  0.2204],
               [ 0.1209,  0.199 ,  0.1453,  4.6505]])
        >>> lag_matrix
        array([[0, 4, 1, 2],
               [0, 0, 0, 0],
               [0, 2, 0, 1],
               [0, 2, 0, 0]], dtype=int8)

        :type tau_max: int [int>=0]
        :arg  tau_max: maximum lag of MI lag function.

        :type knn: int [int>=1]
        :arg  knn: nearest-neighbor MI estimation parameter. (default: 10)

        :type bins: int [int>=2]
        :arg  bins: binning MI estimation parameter. (default: 6)

        :type estimator: str [('knn'|'binning'|'gauss')]
        :arg  estimator: MI estimator. (default: 'knn')

        :type lag_mode: str [('max'|'all')]
        :arg  lag_mode: lag-mode of MI to return.

        :rtype: 3D-array or tuple of matrices
        :returns: all-lag array or matrices of value and lag at the absolute
                  maximum.
        """

        data = self.data
        T, N = data.shape

        # Sanity checks
        if not isinstance(data, numpy.ndarray):
            raise TypeError(f"data is of type {type(data)}, "
                            "must be numpy.ndarray")
        if N > T:
            print(f"Warning: data.shape = {data.shape},"
                  " is it of shape (observations, variables) ?")
        if T < 500:
            print(f"Warning: T = {T} ,"
                  " unreliable estimation using MI estimator")
        assert numpy.isnan(data).sum() == 0, "NaNs in the data"
        assert tau_max >= 0, f"{tau_max =}"
        if estimator == 'knn':
            assert 1 <= knn <= T/2., f"{knn =}"

        if lag_mode == 'max':
            similarity_matrix = numpy.ones((N, N), dtype=FIELD)
            lag_matrix = numpy.zeros((N, N), dtype=LAG)
        elif lag_mode == 'all':
            lagfuncs = numpy.zeros((N, N, tau_max+1), dtype=FIELD)

        if estimator == 'binning':
            self.plogp = self.create_plogp(T)

        for i in range(N):
            for j in range(N):
                maximum = 0.
                lag_at_max = 0
                for tau in range(tau_max + 1):

                    X = [(i, -tau)]
                    Y = [(j, 0)]
                    Z = []

                    XYZ = X + Y + Z
                    dim = len(XYZ)
                    max_lag = tau_max
                    array = numpy.zeros((dim, T - max_lag))
                    for d, node in enumerate(XYZ):
                        var, lag = node
                        array[d, :] = data[max_lag + lag: T + lag, var]

                    if estimator == 'knn':
                        xyz = numpy.array([0, 1])

                        k_xz, k_yz, k_z = self.get_nearest_neighbors(
                            array=array, xyz=xyz, k=knn, standardize=True)

                        ixy_z = (special.digamma(knn)
                                 + (- special.digamma(k_xz)
                                    - special.digamma(k_yz)
                                    + special.digamma(k_z)).mean())

                    elif estimator == 'binning':
                        symb_array = self._quantile_bin_array(array, bins=bins)

                        # High-dimensional Histogram
                        hist = self.bincount_hist(symb_array)

                        # Entropies by use of vectorized function plogp
                        hxyz = (-(self.plogp(hist)).sum()
                                + self.plogp(T))/float(T)
                        hxz = (-(self.plogp(hist.sum(axis=1))).sum()
                               + self.plogp(T))/float(T)
                        hyz = (-(self.plogp(hist.sum(axis=0))).sum()
                               + self.plogp(T))/float(T)
                        hz = (-(self.plogp(hist.sum(axis=0).sum(axis=0))).sum()
                              + self.plogp(T))/float(T)

                        ixy_z = hxz + hyz - hz - hxyz

                    elif estimator == 'gauss':

                        # Standardize
                        array -= array.mean(axis=1).reshape(dim, 1)
                        array /= array.std(axis=1).reshape(dim, 1)
                        if numpy.isnan(array).sum() != 0:
                            raise ValueError("nans after standardizing, \
                                             possibly constant array!")

                        x = array[0, :]
                        y = array[1, :]

                        ixy_z = self._par_corr_to_cmi(
                            numpy.dot(x, y) / numpy.sqrt(
                                numpy.dot(x, x) * numpy.dot(y, y)))

                    if lag_mode == 'max':
                        if ixy_z > maximum:
                            maximum = ixy_z
                            lag_at_max = tau

                    elif lag_mode == 'all':
                        lagfuncs[i, j, tau] = ixy_z

                if lag_mode == 'max':
                    similarity_matrix[i, j] = maximum
                    lag_matrix[i, j] = lag_at_max

        if lag_mode == 'max':
            return similarity_matrix, lag_matrix
        elif lag_mode == 'all':
            return lagfuncs
        else:
            return None

    def information_transfer(self, tau_max=0, estimator='knn',
                             knn=10, past=1, cond_mode='ity', lag_mode='max'):
        r"""
        Return bivariate information transfer between all pairs of nodes.

        Two condition modes of information transfer are available
        as described in [Runge2012b]_.

        Information transfer to Y (ITY):
            .. math::
                I(X^i_t-\tau, X^j_t | X^j_t-1, ...,X^j_t-past)

        Momentary information transfer (MIT):
            .. math::
                I(X^i_t-\tau, X^j_t | X^j_t-1, ...,X^j_t-past, X^i_t-\tau-1,
                                       ...,X^j_t-\tau-past)

        Two estimators are available:

        estimator = 'knn' (Recommended):
        Based on k-nearest-neighbors [Kraskov2004]_,
        version 1 in their paper. Larger k have smaller variance, but larger
        (typically negative) bias, and vice versa.

        estimator = 'gauss':
        Captures only linear part of association. Essentially estimates a
        transformed partial correlation.


        Two lag-modes are available (default: lag_mode='max'):

        lag_mode = 'all':
        Return 3-dimensional array of lag-functions between all pairs of nodes.
        An entry :math:`(i, j, \tau)` corresponds to :math:`I(X^i_t-\tau, X^j_t
        | ...)` for positive lags tau, i.e., the direction i --> j for
        :math:`\tau \ne 0`.

        lag_mode = 'max':
        Return matrix of absolute maxima and corresponding lags of
        lag-functions between all pairs of nodes.
        Returns two usually asymmetric matrices of values and lags: In each
        matrix, an entry :math:`(i, j)` corresponds to the value and lag,
        respectively, at absolute maximum of :math:`I(X^i_t-\tau, X^j_t | ...)`
        for positive lags tau, i.e., the direction i --> j for :math:`\tau >
        0`.  The matrices are, thus, asymmetric. The function
        :meth:`.symmetrize_by_absmax` can be used to obtain a symmetric matrix.

        **Example:**

        >>> coup_ana = CouplingAnalysis(CouplingAnalysis.test_data())
        >>> similarity_matrix, lag_matrix = coup_ana.information_transfer(
        ...     tau_max=5, estimator='knn', knn=10)
        >>> r((similarity_matrix, lag_matrix))
        (array([[ 0.    ,  0.1544,  0.3261,  0.3047],
               [  0.0218,  0.    ,  0.0394,  0.0976],
               [  0.0134,  0.0663,  0.    ,  0.1502],
               [  0.0066,  0.0694,  0.0401,  0.    ]]),
        array([[0, 2, 1, 2], [5, 0, 0, 0], [5, 1, 0, 1], [5, 0, 0, 0]]))

        :type tau_max: int [int>=0]
        :arg  tau_max: maximum lag of ITY lag function.

        :type past: int [int>=1]
        :arg  past: maximum lag of past history.

        :type knn: int [int>=1]
        :arg  knn: nearest-neighbor ITY estimation parameter. (default: 10)

        :type bins: int [int>=2]
        :arg  bins: binning ITY estimation parameter. (default: 6)

        :type estimator: str [('knn'|'gauss')]
        :arg  estimator: ITY estimator. (default: 'knn')

        :type cond_mode: str [('ity'|'mit')]
        :arg  cond_mode: condition mode. (default: 'ity')

        :type lag_mode: str [('max'|'all')]
        :arg  lag_mode: lag-mode of ITY to return.

        :rtype: 3D-array or tuple of matrices
        :returns: all-lag array or matrices of value and lag at the absolute
                  maximum.
        """

        data = self.data
        T, N = data.shape

        # Sanity checks
        if not isinstance(data, numpy.ndarray):
            raise TypeError(f"data is of type {type(data)},"
                            " must be numpy.ndarray")
        if N > T:
            print(f"Warning: data.shape = {data.shape},"
                  " is it of shape (observations, variables) ?")
        if estimator == 'knn' and T < 500:
            print(f"Warning: T = {T} ,"
                  " unreliable estimation using knn-estimator")
        if numpy.isnan(data).sum() != 0:
            raise ValueError("NaNs in the data")
        if tau_max < 0:
            raise ValueError(f"tau_max = {tau_max}, but 0 <= tau_max")
        if estimator == 'knn':
            if knn > T/2. or knn < 1:
                raise ValueError(f"knn = {knn}, should be between 1 and T/2")

        if lag_mode == 'max':
            similarity_matrix = numpy.ones((N, N), dtype=FIELD)
            lag_matrix = numpy.zeros((N, N), dtype=LAG)
        elif lag_mode == 'all':
            lagfuncs = numpy.zeros((N, N, tau_max+1), dtype=FIELD)

        for i in range(N):
            for j in range(N):
                maximum = 0.
                lag_at_max = 0
                for tau in range(tau_max + 1):

                    X = [(i, -tau)]
                    Y = [(j, 0)]
                    if cond_mode == 'ity':
                        Z = [(j, -p) for p in range(1, past + 1)]
                    elif cond_mode == 'mit':
                        Z = [(j, -p) for p in range(1, past + 1)]
                        Z += [(i, -tau - p) for p in range(1, past + 1)]

                    XYZ = X + Y + Z

                    dim = len(XYZ)
                    max_lag = tau_max + past
                    array = numpy.zeros((dim, T - max_lag))
                    for d, node in enumerate(XYZ):
                        var, lag = node
                        array[d, :] = data[max_lag + lag: T + lag, var]

                    if estimator == 'knn':
                        xyz = numpy.array([0, 1])

                        k_xz, k_yz, k_z = self.get_nearest_neighbors(
                            array=array, xyz=xyz, k=knn, standardize=True)

                        ixy_z = (special.digamma(knn)
                                 + (- special.digamma(k_xz)
                                    - special.digamma(k_yz)
                                    + special.digamma(k_z)).mean())

                    elif estimator == 'gauss':

                        if numpy.isnan(array).sum() != 0:
                            raise ValueError("nans in the array!")

                        # Standardize
                        array -= array.mean(axis=1).reshape(dim, 1)
                        array /= array.std(axis=1).reshape(dim, 1)
                        if numpy.isnan(array).sum() != 0:
                            raise ValueError("nans after standardizing, \
                                             possibly constant array!")

                        x = array[0, :]
                        y = array[1, :]
                        if len(array) > 2:
                            confounds = array[2:, :]
                            ortho_confounds = linalg.qr(confounds.T.copy(),
                                                        mode='economic')[0].T
                            x -= numpy.dot(numpy.dot(ortho_confounds, x),
                                           ortho_confounds)
                            y -= numpy.dot(numpy.dot(ortho_confounds, y),
                                           ortho_confounds)

                        ixy_z = self._par_corr_to_cmi(
                            numpy.dot(x, y) / numpy.sqrt(
                                numpy.dot(x, x) * numpy.dot(y, y)))

                    if lag_mode == 'max':
                        if ixy_z > maximum:
                            maximum = ixy_z
                            lag_at_max = tau

                    elif lag_mode == 'all':
                        lagfuncs[i, j, tau] = ixy_z

                if lag_mode == 'max':
                    similarity_matrix[i, j] = maximum
                    lag_matrix[i, j] = lag_at_max

        if lag_mode == 'max':
            similarity_matrix[range(N), range(N)] = 0.
        elif lag_mode == 'all':
            lagfuncs[range(N), range(N), 0.] = 0.

        if lag_mode == 'max':
            return similarity_matrix, lag_matrix
        elif lag_mode == 'all':
            return lagfuncs
        else:
            return None

    #
    #  Define helper methods
    #

    @staticmethod
    def _par_corr_to_cmi(par_corr):
        """
        Transformation of partial correlation to conditional mutual
        information scale using the (multivariate) Gaussian assumption.

        :type par_corr: float or array
        :arg  par_corr: partial correlation

        :rtype: float
        :returns: transformed partial correlation.
        """

        return -0.5*numpy.log(1. - par_corr**2)

    @staticmethod
    def get_nearest_neighbors(array, xyz, k, standardize=True):
        """
        Returns nearest-neighbors for conditional mutual information estimator.

        Reference: [Kraskov2004]_

        :type array: array (float)
        :arg  array: data array.

        :type xyz: array [int(0|1|2)]
        :arg  xyz: identifier of X, Y, Z in CMI

        :type k: int [int>=1]
        :arg  k: nearest-neighbor MI estimation parameter.

        :type standardize: bool
        :arg  standardize: standardize array before estimation. (default: True)

        :rtype: tuple of arrays
        :returns: nearest neighbors for each sample point.
        """
        dim, T = array.shape

        if standardize:
            # Standardize
            array = array.astype(FIELD)
            array -= array.mean(axis=1).reshape(dim, 1)
            array /= array.std(axis=1).reshape(dim, 1)
            # If the time series is constant, return nan rather than raising
            # Exception
            if numpy.isnan(array).sum() != 0:
                raise ValueError("nans after standardizing, possibly \
                                 constant array!")

        # Add noise to destroy ties...
        array += 1E-10 * numpy.random.rand(dim, T)

        dim_x = int(numpy.where(xyz == 0)[0][-1] + 1)
        dim_y = int(numpy.where(xyz == 1)[0][-1] + 1 - dim_x)
        # dim_z = maxdim - dim_x - dim_y

        return _get_nearest_neighbors(
            to_cy(array, FIELD), dim, T, dim_x, dim_y, k)

    @staticmethod
    def _quantile_bin_array(array, bins=6):
        """
        Returns symbolified array with aequi-quantile binning.

        This partition results in a uniform distribution of the marginals.

        :type array: array
        :arg array: data

        :type bins: int
        :arg bins: number of bins

        :rtype: array
        :returns: converted data
        """

        dim, T = array.shape

        # get the bin quantile steps
        bin_edge = numpy.ceil(T/float(bins))

        symb_array = numpy.zeros((dim, T), dtype=INT32TYPE)

        # get the lower edges of the bins for every time series
        edges = numpy.sort(array, axis=1)[:, ::bin_edge]
        bins = edges.shape[1]

        # This gives the symbolic time series
        symb_array = (array.reshape(dim, T, 1)
                      >= edges.reshape(dim, 1, bins)).sum(axis=2) - 1

        return symb_array

    @staticmethod
    def bincount_hist(symb_array):
        """
        Computes histogram from symbolic array.

        :type symb_array: array of integers
        :arg symb_array: symbolic data

        :rtype: array
        :returns: (unnormalized) histogram
        """

        base = int(symb_array.max() + 1)

        D, T = symb_array.shape

        # Needed because numpy.bincount cannot process longs
        assert isinstance(base**D, int)
        assert base**D*16./8./1024.**3 < 3., (
            'Dimension exceeds 3 GB of necessary memory '
            '(change this code line if you got more...)')
        assert D*base**D < 2**65, (
            f'base = {base}, D = {D}: Histogram failed:'
            ' dimension D*base**D exceeds int64 data type')

        flathist = numpy.zeros((base**D), dtype=INT16TYPE)
        multisymb = numpy.zeros(T, dtype=INT64TYPE)

        for i in range(D):
            multisymb += symb_array[i, :]*base**i

        result = numpy.bincount(multisymb)
        flathist[:len(result)] += result

        return flathist.reshape(tuple(
            [base, base] + [base for i in range(D-2)])).T

    @staticmethod
    def create_plogp(T):
        """
        Precalculation of p*log(p) needed for entropies.

        :type T: int
        :arg  T: sample length

        :rtype: array
        :returns: p*log(p) array from p=1 to p=T
        """

        gfunc = numpy.zeros(T+1)
        gfunc[1:] = numpy.arange(1, T+1, 1)*numpy.log(numpy.arange(1, T+1, 1))

        return numpy.vectorize(lambda t: gfunc[t])



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
funcnet
=======

Related Publications
~~~~~~~~~~~~~~~~~~~~

"""

from .coupling_analysis import CouplingAnalysis
from .coupling_analysis_pure_python import CouplingAnalysisPurePython



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for the analysis of dynamical systems and time series based
on recurrence plots, including measures of recurrence quantification
analysis (RQA) and recurrence network analysis.
"""

# array object and fast numerics
import numpy as np

from ..core import InteractingNetworks

from ..core._ext.types import to_cy, ADJ, MASK, FIELD
from ._ext.numerics import _visibility_relations_missingvalues, \
    _visibility_relations_no_missingvalues, _visibility_relations_horizontal, \
    _retarded_local_clustering, _advanced_local_clustering

#
#  Class definitions
#


class VisibilityGraph(InteractingNetworks):
    """
    Class VisibilityGraph for generating and analyzing visibility graphs of
    time series.

    Visibility graphs were initially applied for time series analysis by
    [Lacasa2008]_.
    """

    #
    #  Internal methods
    #

    def __init__(self, time_series, timings=None, missing_values=False,
                 horizontal=False, silence_level=0):
        """
        Missing values are handled as infinite values, effectively separating
        the visibility graph into different disconnected components.

        .. note::
           Missing values have to be marked by the Numpy NaN flag!

        :type time_series: 2D array (time, dimension)
        :arg time_series: The time series to be analyzed, can be scalar or
            multi-dimensional.
        :arg str timings: Timings of the observations in :attr:`time_series`.
        :arg bool missing_values: Toggle special treatment of missing values in
            :attr:`time_series`.
        :arg bool horizontal: Indicates whether a horizontal visibility
            relation is used.
        :arg number silence_level: Inverse level of verbosity of the object.
        """
        #  Set silence_level
        self.silence_level = silence_level
        """The inverse level of verbosity of the object."""

        #  Set missing_values flag
        self.missing_values = missing_values
        """Controls special treatment of missing values in
           :attr:`time_series`."""

        #  Store time series
        self.time_series = to_cy(time_series, FIELD)
        """The time series from which the visibility graph is constructed."""

        if timings is not None:
            timings = to_cy(timings, FIELD)
        else:
            timings = np.arange(len(time_series), dtype=FIELD)

        #  Store timings
        self.timings = timings
        """The timimgs of the time series data points."""

        #  Get missing value indices
        if self.missing_values:
            self.missing_value_indices = np.isnan(self.time_series)

        #  Determine visibility relations
        if not horizontal:
            A = self.visibility_relations()
        else:
            A = self.visibility_relations_horizontal()

        #  Initialize Network object
        InteractingNetworks.__init__(self, A, directed=False,
                                     silence_level=silence_level)

    def __str__(self):
        """
        Returns a string representation.
        """
        return ("VisibilityGraph: "
                f"time series shape {self.time_series.shape}.\n"
                f"{InteractingNetworks.__str__(self)}")

    #
    #  Visibility methods
    #

    def visibility_relations(self):
        """
        Returns visibility between all nodes of self.timeseries
        :rtype: 2D array of MASK
        """
        if self.silence_level <= 1:
            print("Calculating visibility relations...")

        #  Prepare
        x = self.time_series
        t = self.timings
        N = len(self.time_series)
        A = np.zeros((N, N), dtype=MASK)

        if self.missing_values:
            mv_indices = self.missing_value_indices
            _visibility_relations_missingvalues(x, t, N, A, mv_indices)

        else:
            _visibility_relations_no_missingvalues(x, t, N, A)

        return A

    # FIXME: There is no option for missing values
    def visibility_relations_horizontal(self):
        """
        Returns horizontal visibility between all nodes of self.timeseries
        :rtype: 2D array of MASK
        """
        if self.silence_level <= 1:
            print("Calculating horizontal visibility relations...")

        #  Prepare
        x = self.time_series
        N = len(self.time_series)
        A = np.zeros((N, N), dtype=MASK)

        _visibility_relations_horizontal(x, N, A)
        return A

    #
    #  Specific measures for visibility graphs
    #

    def visibility(self, node1, node2):
        """
        Returns the visibility between node 1 and 2 as boolean.
        :arg int node1: node index of node 1
        :arg int node2: node index of node 2
        :rtype: bool
        """
        return self.adjacency[node1, node2]

    def visibility_single(self, node):
        """
        Returns the visibility between all nodes of self.time_series and node
        as array of booleans.
        :arg int node: node index of the node
        :rtype: 1D array of bool
        """
        return self.adjacency[node, :]

    def retarded_degree(self):
        """Return number of neighbors in the past of a node."""
        #  Prepare
        retarded_degree = np.zeros(self.N)
        A = self.adjacency

        for i in range(self.N):
            retarded_degree[i] = A[i, :i].sum()

        return retarded_degree

    def advanced_degree(self):
        """Return number of neighbors in the future of a node."""
        #  Prepare
        advanced_degree = np.zeros(self.N)
        A = self.adjacency

        for i in range(self.N):
            advanced_degree[i] = A[i, i:].sum()

        return advanced_degree

    def retarded_local_clustering(self):
        """
        Return probability that two neighbors of a node in its past are
        connected.
        """
        #  Prepare
        retarded_clustering = np.zeros(self.N)

        #  Get full adjacency matrix
        A = self.adjacency
        #  Get number of nodes
        N = self.N

        #  Get left degree
        retarded_degree = self.retarded_degree()
        #  Prepare normalization factor
        norm = retarded_degree * (retarded_degree - 1) / 2.

        _retarded_local_clustering(N, to_cy(A, ADJ), norm, retarded_clustering)
        return retarded_clustering

    def advanced_local_clustering(self):
        """
        Return probability that two neighbors of a node in its future are
        connected.
        """
        #  Prepare
        advanced_clustering = np.zeros(self.N)

        #  Get full adjacency matrix
        A = self.adjacency
        #  Get number of nodes
        N = self.N

        #  Get right degree
        advanced_degree = self.advanced_degree()
        #  Prepare normalization factor
        norm = advanced_degree * (advanced_degree - 1) / 2.

        _advanced_local_clustering(N, to_cy(A, ADJ), norm, advanced_clustering)
        return advanced_clustering

    def retarded_closeness(self):
        """Return average path length to nodes in the past of a node."""
        #  Prepare
        retarded_closeness = np.zeros(self.N)
        path_lengths = self.path_lengths()

        for i in range(self.N):
            retarded_closeness[i] = path_lengths[i, :i].mean() ** (-1)

        return retarded_closeness

    def advanced_closeness(self):
        """Return average path length to nodes in the future of a node."""
        #  Prepare
        advanced_closeness = np.zeros(self.N)
        path_lengths = self.path_lengths()

        for i in range(self.N):
            advanced_closeness[i] = path_lengths[i, i+1:].mean() ** (-1)

        return advanced_closeness

    def retarded_betweenness(self):
        """
        Return betweenness of a node with respect to all pairs of nodes in its
        past.
        """
        #  Prepare
        retarded_betweenness = np.zeros(self.N)

        for i in range(self.N):
            retarded_indices = np.arange(i)
            retarded_betweenness[i] = self.nsi_betweenness(
                sources=retarded_indices, targets=retarded_indices)[i]

        return retarded_betweenness

    def advanced_betweenness(self):
        """
        Return betweenness of a node with respect to all pairs of nodes in its
        future.
        """
        #  Prepare
        advanced_betweenness = np.zeros(self.N)

        for i in range(self.N):
            advanced_indices = np.arange(i+1, self.N)
            advanced_betweenness[i] = self.nsi_betweenness(
                sources=advanced_indices, targets=advanced_indices)[i]

        return advanced_betweenness

    def trans_betweenness(self):
        """
        Return betweenness of a node with respect to all pairs of nodes
        with one node the past and one node in the future, respectively.
        """
        #  Prepare
        trans_betweenness = np.zeros(self.N)

        for i in range(self.N):
            retarded_indices = np.arange(i)
            advanced_indices = np.arange(i+1, self.N)
            trans_betweenness[i] = self.nsi_betweenness(
                sources=retarded_indices, targets=advanced_indices)[i]

        return trans_betweenness

    #
    #  Measures corrected for boundary effects
    #

    def boundary_corrected_degree(self):
        """Return a weighted degree corrected for trivial boundary effects."""
        #  Prepare
        N_past = np.arange(self.N)
        N_future = N_past[::-1]

        cdegree = (self.retarded_degree() * N_past
                   + self.advanced_degree() * N_future) / float(self.N - 1)

        return cdegree

    def boundary_corrected_closeness(self):
        """
        Return a weighted closeness corrected for trivial boundary effects.
        """
        #  Prepare
        N_past = np.arange(self.N)
        N_future = N_past[::-1]

        ccloseness = (self.N - 1) * (self.retarded_closeness() / N_past
                                     + self.advanced_closeness() / N_future)

        return ccloseness



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

cimport cython
cimport numpy as cnp

cnp.import_array()


ctypedef cnp.int8_t BOOLTYPE_t
ctypedef cnp.int8_t INT8TYPE_t
ctypedef cnp.int16_t INT16TYPE_t
ctypedef cnp.int32_t INT32TYPE_t
ctypedef cnp.int64_t INT64TYPE_t
ctypedef cnp.float32_t FLOAT32TYPE_t
ctypedef cnp.float64_t FLOAT64TYPE_t

ctypedef BOOLTYPE_t ADJ_t
ctypedef BOOLTYPE_t MASK_t
ctypedef INT8TYPE_t LAG_t
ctypedef INT16TYPE_t DEGREE_t
ctypedef INT32TYPE_t NODE_t
ctypedef FLOAT32TYPE_t WEIGHT_t
ctypedef FLOAT64TYPE_t DWEIGHT_t
ctypedef FLOAT32TYPE_t FIELD_t
ctypedef FLOAT64TYPE_t DFIELD_t



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

from pathlib import Path

import requests
import pytest


@pytest.fixture(scope="session",
                params=["supremum", "euclidean", "manhattan"])
def metric(request) -> str:
    '''
    A fixture for creating parametrized fixtures of classes that have a
    `metric` argument, as in `RecurrencePlot` and its child classes.
    '''
    return request.param


@pytest.fixture(scope="session")
def reanalysis_data() -> Path:
    """
    Locate, and potentially download, a small NOAA dataset. Currently used in:
    - `tests/test_climate/test_map_plot.py`
    - `examples/tutorials/ClimateNetworks.ipynb`
    """
    service = "https://psl.noaa.gov/repository/entry/get"
    data_name = "air.mon.mean.nc"
    query = (
        "entryid=synth%3Ae570c8f9-ec09-4e89-93b4-"
        "babd5651e7a9%3AL25jZXAucmVhbmFseXNpcy5kZXJpdm"
        "VkL3N1cmZhY2UvYWlyLm1vbi5tZWFuLm5j")
    url = f"{service}/{data_name}?{query}"

    data_dir = Path("./examples/tutorials/data")
    data_file = data_dir / data_name

    if not data_dir.is_dir():
        data_dir.mkdir(parents=True)
    if not data_file.exists():
        res = requests.get(url, timeout=(30, 30))
        res.raise_for_status()
        with open(data_file, 'wb') as f:
            f.write(res.content)

    return data_file



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex climate networks.
"""

#
#  Import essential packages
#

#  Import NumPy for the array object and fast numerics
import numpy as np

from .tsonis import TsonisClimateNetwork


#
#  Define class PartialCorrelationClimateNetwork
#

class PartialCorrelationClimateNetwork(TsonisClimateNetwork):

    """
    Encapsulates a partial correlation climate network.

    Constructs a static climate network based on partial correlation, as in
    [Ueoka2008]_.
    """

    #
    #  Defines internal methods
    #

    def __init__(self, data, threshold=None, link_density=None,
                 non_local=False, node_weight_type="surface", winter_only=True,
                 silence_level=0):
        """
        Initialize an instance of PartialCorrelationClimateNetwork.

        .. note::
           Either threshold **OR** link_density have to be given!

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos**2 lat)

        :type data: :class:`.ClimateData`
        :arg data: The climate data used for network construction.
        :arg float threshold: The threshold of similarity measure, above which
            two nodes are linked in the network.
        :arg float link_density: The networks's desired link density.
        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        :arg str node_weight_type: The type of geographical node weight to be
            used.
        :arg bool winter_only: Determines, whether only data points from the
            winter months (December, January and February) should be used for
            analysis. Possibly, this further suppresses the annual cycle in the
            time series.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if silence_level <= 1:
            print("Generating a partial correlation climate network...")

        #  Call constructor of parent class TsonisClimateNetwork
        TsonisClimateNetwork.__init__(self, data=data, threshold=threshold,
                                      link_density=link_density,
                                      non_local=non_local,
                                      node_weight_type=node_weight_type,
                                      winter_only=winter_only,
                                      silence_level=silence_level)

    def __str__(self):
        """
        Return a string representation of PartialCorrelationClimateNetwork.
        """
        return ('PartialCorrelationClimateNetwork:\n'
                f'{TsonisClimateNetwork.__str__(self)}')

    #
    #  Defines methods to calculate the correlation matrix
    #

    def _calculate_correlation(self, anomaly):
        """
        Return the partial correlation matrix at zero lag.

        :type anomaly: 2D Numpy array (time, index)
        :arg anomaly: the anomaly time series from to calculate the partial
                      correlation matrix at zero lag.

        :rtype: 2D Numpy array (index, index)
        :return: the partial correlation matrix at zero lag.
        """
        if self.silence_level <= 1:
            print("Calculating partial correlation matrix at zero lag from "
                  "anomaly values...")

        #  Calculate the correlation matrix, cast to float64 for precise
        #  calculation of inverse matrix.
        C = np.corrcoef(anomaly.transpose()).astype("float64")

        #  Calculate the inverse correlation matrix
        if np.linalg.det(C) != 0.0:
            C_inv = np.linalg.inv(C)
        else:
            C_inv = np.linalg.pinv(C)

        #  Clean up
        del C

        #  Get the diagonal of the inverse correlation matrix
        diag = C_inv.diagonal()[:]

        #  Calculate matrix of normalizations
        norm = np.sqrt(abs(np.outer(diag, diag)))

        return - C_inv / norm



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"
"""
Simple tests for the ClimateData class.
"""
import numpy as np

from pyunicorn.core.data import Data
from pyunicorn.climate.climate_data import ClimateData

# -----------------------------------------------------------------------------
# Class member tests
# -----------------------------------------------------------------------------


def test_SmallTestData():
    res = Data.SmallTestData().observable()
    exp = np.array([[0., 1., 0., -1., -0., 1.],
                    [0.309, 0.9511, -0.309, -0.9511, 0.309, 0.9511],
                    [0.5878, 0.809, -0.5878, -0.809, 0.5878, 0.809],
                    [0.809, 0.5878, -0.809, -0.5878, 0.809, 0.5878],
                    [0.9511, 0.309, -0.9511, -0.309, 0.9511, 0.309],
                    [1., 0., -1., -0., 1., 0.],
                    [0.9511, -0.309, -0.9511, 0.309, 0.9511, -0.309],
                    [0.809, -0.5878, -0.809, 0.5878, 0.809, -0.5878],
                    [0.5878, -0.809, -0.5878, 0.809, 0.5878, -0.809],
                    [0.309, -0.9511, -0.309, 0.9511, 0.309, -0.9511]])
    assert np.allclose(res, exp, atol=1e-04)


def test_phase_indices():
    res = ClimateData.SmallTestData().phase_indices()
    exp = np.array([[0, 5], [1, 6], [2, 7], [3, 8], [4, 9]])
    assert (res == exp).all()


def test_indices_selected_phases():
    res = ClimateData.SmallTestData().indices_selected_phases([0, 1, 4])
    exp = np.array([0, 1, 4, 5, 6, 9])
    assert (res == exp).all()


def test_phase_mean():
    res = ClimateData.SmallTestData().phase_mean()
    exp = np.array([[0.5, 0.5, -0.5, -0.5, 0.5, 0.5],
                    [0.63, 0.321, -0.63, -0.321, 0.63, 0.321],
                    [0.6984, 0.1106, -0.6984, -0.1106, 0.6984, 0.1106],
                    [0.6984, -0.1106, -0.6984, 0.1106, 0.6984, -0.1106],
                    [0.63, -0.321, -0.63, 0.321, 0.63, -0.321]])
    assert np.allclose(res, exp, atol=1e-04)


def test_anomaly():
    res = ClimateData.SmallTestData().anomaly()[:, 0]
    exp = np.array([-0.5, -0.321, -0.1106, 0.1106, 0.321,
                    0.5, 0.321, 0.1106, -0.1106, -0.321])
    assert np.allclose(res, exp, atol=1e-04)


def test_shuffled_anomaly():
    res = ClimateData.SmallTestData().anomaly().std(axis=0)
    exp = np.array([0.31, 0.6355, 0.31, 0.6355, 0.31, 0.6355])
    assert np.allclose(res, exp, atol=1e-04)

    res = ClimateData.SmallTestData().shuffled_anomaly().std(axis=0)
    exp = np.array([0.31, 0.6355, 0.31, 0.6355, 0.31, 0.6355])
    assert np.allclose(res, exp, atol=1e-04)


def test_window():
    data = ClimateData.SmallTestData()
    data.set_window(window={"time_min": 0., "time_max": 0.,
                            "lat_min": 10., "lat_max": 20.,
                            "lon_min": 5., "lon_max": 10.})
    res = data.anomaly()
    exp = np.array([[0.5, -0.5], [0.321, -0.63], [0.1106, -0.6984],
                    [-0.1106, -0.6984], [-0.321, -0.63], [-0.5, 0.5],
                    [-0.321, 0.63], [-0.1106, 0.6984], [0.1106, 0.6984],
                    [0.321, 0.63]])
    assert np.allclose(res, exp, atol=1e-04)


def test_set_global_window():
    data = ClimateData.SmallTestData()
    data.set_window(window={"time_min": 0., "time_max": 4.,
                            "lat_min": 10., "lat_max": 20.,
                            "lon_min": 5., "lon_max": 10.})
    res = data.grid.grid()["lat"]
    exp = np.array([10., 15.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)

    data.set_global_window()
    res = data.grid.grid()["lat"]
    exp = np.array([0., 5., 10., 15., 20., 25.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for the analysis of dynamical systems and time series based
on recurrence plots, including measures of recurrence quantification
analysis (RQA) and recurrence network analysis.
"""

from math import factorial
from typing import Tuple
from collections.abc import Hashable

import numpy as np
from numpy.typing import NDArray

from ..core.cache import Cached
from ..core._ext.types import to_cy, NODE, LAG, FIELD, DFIELD
from ._ext.numerics import _embed_time_series, _manhattan_distance_matrix_rp, \
    _euclidean_distance_matrix_rp, _supremum_distance_matrix_rp, \
    _set_adaptive_neighborhood_size, _bootstrap_distance_matrix_manhattan, \
    _bootstrap_distance_matrix_euclidean, \
    _bootstrap_distance_matrix_supremum, \
    _diagline_dist_missingvalues, _diagline_dist, \
    _diagline_dist_sequential_missingvalues, _diagline_dist_sequential, \
    _vertline_dist_missingvalues, _vertline_dist, \
    _vertline_dist_sequential_missingvalues, _vertline_dist_sequential, \
    _rejection_sampling, _white_vertline_dist, _twins_r, _twin_surrogates_r


class RecurrencePlot(Cached):
    """
    Class RecurrencePlot for generating and quantitatively analyzing
    :index:`recurrence plots <single: recurrence plot>`.

    The RecurrencePlot class supports the construction of recurrence plots
    from multi-dimensional time series, optionally using embedding. Currently,
    manhattan, euclidean and supremum norms are provided for measuring
    distances in phase space.

    Methods for calculating commonly used measures of :index:`recurrence
    quantification analysis <pair: RQA; recurrence plot>` (RQA) are provided,
    e.g., determinism, maximum diagonal line length and laminarity. The
    definitions of these measures together with a review of the theory and
    applications of recurrence plots are given in [Marwan2007]_.

    **Examples:**

     - Create an instance of RecurrencePlot with a fixed recurrence threshold
       and without embedding::

           RecurrencePlot(time_series, threshold=0.1)

     - Create an instance of RecurrencePlot with a fixed recurrence threshold
       in units of STD and without embedding::

           RecurrencePlot(time_series, threshold_std=0.03)

     - Create an instance of RecurrencePlot at a fixed (global) recurrence rate
       and using time delay embedding::

           RecurrencePlot(time_series, dim=3, tau=2,
                          recurrence_rate=0.05).recurrence_rate()
    """

    #
    #  Internal methods
    #

    def __init__(self, time_series: NDArray, metric: str = "supremum",
                 normalize: bool = False, missing_values: bool = False,
                 sparse_rqa: bool = False, silence_level: int = 0,
                 **kwargs):
        """
        Initialize an instance of RecurrencePlot.

        Either recurrence threshold ``threshold``/``threshold_std``, recurrence
        rate ``recurrence_rate`` or local recurrence rate
        ``local_recurrence_rate`` have to be given as keyword arguments.

        Embedding is only supported for scalar time series. If embedding
        dimension ``dim`` and delay ``tau`` are **both** given as keyword
        arguments, embedding is applied. Multidimensional time series are
        processed as is by default.

        :attention: The sparse_rqa feature is experimental and currently only
                    works for fixed threshold and the supremum metric.

        :type time_series: 2D array (time, dimension)
        :arg time_series: The time series to be analyzed, can be scalar or
            multi-dimensional.
        :arg str metric: The metric for measuring distances in phase space
            ("manhattan", "euclidean", "supremum").
        :arg bool normalize: Decide whether to normalize the time series to
            zero mean and unit standard deviation.
        :arg bool missing_values: Toggle special treatment of missing values in
            :attr:`.RecurrencePlot.time_series`.
        :arg bool sparse_rqa: Toggles sequential RQA computation using less
            memory for use with long time series.
        :arg bool skip_recurrence: Skip calculation of recurrence matrix within
            RP class (e.g. when overloading respective methods in child class)
        :arg int silence_level: Inverse level of verbosity of the object.
        :arg number threshold: The recurrence threshold keyword for generating
            the recurrence plot using a fixed threshold.
        :arg number threshold_std: The recurrence threshold keyword for
            generating the recurrence plot using a fixed threshold in units of
            the time series' STD.
        :arg number recurrence_rate: The recurrence rate keyword for generating
            the recurrence plot using a fixed recurrence rate.
        :arg number local_recurrence_rate: The local recurrence rate keyword
            for generating the recurrence plot using a fixed local recurrence
            rate (same number of recurrences for each state vector).
        :arg number adaptive_neighborhood_size: The adaptive neighborhood size
            parameter for generating recurrence plots based on the algorithm in
            [Xu2008]_.
        :arg number dim: The embedding dimension.
        :arg number tau: The embedding delay.
        """
        #  Set silence_level
        self.silence_level = silence_level
        """The inverse level of verbosity of the object."""

        #  Set missing_values flag
        self.missing_values = missing_values
        """Controls special treatment of missing values in
           :attr:`.RecurrencePlot.time_series`."""

        #  Set sparse RQA flag
        self.sparse_rqa = sparse_rqa
        """Controls sequential calculation of RQA measures."""

        #  Store time series
        self.time_series = to_cy(time_series, FIELD)
        """The time series from which the recurrence plot is constructed."""

        #  Reshape time series
        self.time_series.shape = (self.time_series.shape[0], -1)

        #  Store type of metric
        self._known_metrics = ("manhattan", "euclidean", "supremum")
        assert metric in self._known_metrics
        self.metric = metric
        """The metric used for measuring distances in phase space."""

        # minimal denominator for numerical stability
        self._epsilon = 1e-08

        #  Normalize time series
        if normalize:
            self.normalize_time_series(self.time_series)

        #  Get embedding dimension and delay from **kwargs
        self.dim = kwargs.get("dim")
        self.tau = kwargs.get("tau")

        self.N: int = 0
        """The number of state vectors (number of lines and rows) of the RP."""
        self.R = None
        """The recurrence matrix."""

        self._mut_embedding: int = 0
        if (self.dim is not None) and (self.tau is not None):
            #  Embed the time series
            self.embedding = self.embed_time_series(
                self.time_series, self.dim, self.tau)
        else:
            self.embedding = self.time_series

        #  Get missing value indices
        if self.missing_values:
            self.missing_value_indices = \
                np.isnan(self.embedding).sum(axis=1) != 0

        #  Get threshold or recurrence rate from **kwargs, construct recurrence
        #  plot accordingly
        self.threshold = kwargs.get("threshold")
        self.threshold_std = kwargs.get("threshold_std")
        #  Make sure not to overwrite the method recurrence_rate()
        recurrence_rate = kwargs.get("recurrence_rate")
        self.local_recurrence_rate = kwargs.get("local_recurrence_rate")
        self.adaptive_neighborhood_size = \
            kwargs.get("adaptive_neighborhood_size")

        #  Precompute recurrence matrix only if sequential RQA is switched off,
        #  and not calling from child class with respective overriding methods.
        skip_recurrence = kwargs.get("skip_recurrence")

        if not sparse_rqa and not skip_recurrence:
            if self.threshold is not None:
                #  Calculate the recurrence matrix R using the radius of
                #  neighborhood threshold
                RecurrencePlot.set_fixed_threshold(self, self.threshold)
            elif self.threshold_std is not None:
                #  Calculate the recurrence matrix R using the radius of
                #  neighborhood threshold in units of the time series' STD
                RecurrencePlot.set_fixed_threshold_std(
                    self, self.threshold_std)
            elif recurrence_rate is not None:
                #  Calculate the recurrence matrix R using a fixed recurrence
                #  rate
                RecurrencePlot.set_fixed_recurrence_rate(
                    self, recurrence_rate)
            elif self.local_recurrence_rate is not None:
                #  Calculate the recurrence matrix R using a fixed local
                #  recurrence rate
                RecurrencePlot.set_fixed_local_recurrence_rate(
                    self, self.local_recurrence_rate)
            elif self.adaptive_neighborhood_size is not None:
                #  Calculate the recurrence matrix R using the adaptive
                #  neighborhood size algorithm in [Xu2008]_
                RecurrencePlot.set_adaptive_neighborhood_size(
                    self, self.adaptive_neighborhood_size)
            else:
                raise NameError("Please give either threshold or \
                                recurrence_rate to construct the recurrence \
                                plot!")

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        return (self._mut_embedding,)

    def __str__(self):
        """
        Returns a string representation.
        """
        return ("RecurrencePlot: "
                f"time series shape {self.time_series.shape}.\n"
                f"Embedding dimension {self.dim if self.dim else 0}\n"
                f"Threshold {self.threshold}, {self.metric} metric")

    @property
    def embedding(self) -> np.ndarray:
        """
        The embedded time series / phase space trajectory
        (time, embedding dimension).
        """
        return self._embedding

    @embedding.setter
    def embedding(self, embedding: np.ndarray):
        self._embedding = to_cy(embedding, DFIELD)
        self.N = self._embedding.shape[0]
        self._mut_embedding += 1

    #
    #  Service methods
    #

    def recurrence_matrix(self):
        """
        Return the current recurrence matrix :math:`R`.

        :rtype: 2D square Numpy array
        :return: the current recurrence matrix :math:`R`.
        """
        if not self.sparse_rqa:
            return self.R
        else:
            print("Exception: Sequential RQA mode is enabled. "
                  "Recurrence matrix is not stored in memory.")
            return None

    def distance_matrix(self, metric: str):
        """
        Return phase space distance matrix :math:`D` according to the chosen
        metric.

        :arg str metric: The metric for measuring distances in phase space
            ("manhattan", "euclidean", "supremum").
        :rtype: 2D square array
        :return: the phase space distance matrix :math:`D`
        """
        assert metric in self._known_metrics, f"unknown metric: {metric}"
        return getattr(RecurrencePlot, f"{metric}_distance_matrix")(self)

    #
    #  Time series handling methods
    #

    @staticmethod
    def normalize_time_series(time_series):
        """
        :index:`Normalize <pair: normalize; time series>` each component of a
        time series **in place**.

        Works also for complex valued time series.

        .. note::
           Modifies the given array in place!

        :type time_series: 2D array (time, dimension)
        :arg time_series: The time series to be normalized.
        """
        #  Get number of components of time_series
        dim = time_series.shape[1]

        mean = time_series.mean(axis=0)
        std = time_series.std(axis=0)

        #  Normalize all components separately
        for i in range(dim):
            time_series[:, i] -= mean[i]
            if std[i] != 0:
                time_series[:, i] /= std[i]

    # Heitzig:
    @staticmethod
    def legendre_coordinates(x, dim=3, t=None, p=None, tau_w="est"):
        """
        Return a phase space trajectory reconstructed using orthogonal
        polynomial filters.

        The reconstructed state vector components are the zero-th to (dim-1)-th
        derivatives of the (possibly irregularly spaced) time series x
        as estimated by folding with the orthogonal polynomial filters that
        correspond to the sequence of measurement time points t.

        This is a generalization for irregularly spaced time series
        of the "Legendre coordinates" introduced in Gibson et al. (1992).

        :arg array-like x: Time series values
        :arg int dim: Dimension > 0 of reconstructed phase space. Default: 3
        :type t: array-like or None
        :arg t: Optional array of measurement time points corresponding to the
            values in x. Default: [0,...,x.size-1]
        :type p: int > 0 or None
        :arg p: No. of past and future time points to use for the estimation.
            Default: dim or determined by tau_w if given
        :type tau_w: float > 0 or "est" or None
        :arg tau_w: Optional (average) window width to use in determining p
            when p = None. Following Gibson et al. (1992), this should be about
            sqrt(3/< x**2 >) * std(x), or about a quarter period.  If "est",
            this is estimated iteratively, starting with
            4 * (max(t)-min(t)) / (N-1) and estimating x' from that.
        :rtype:  2D array [observation index, dimension index]
        :return: Estimated derivatives. Rows are reconstructed state vectors.
        """
        x = np.array(x).flatten()
        N = x.size

        # time points:
        if t is None:
            t = np.arange(N)

        if p is None:
            if tau_w == "est":
                tau_w = 4 * (t.max() - t.min()) / (N-1)
                for i in range(5):
                    y0 = RecurrencePlot.legendre_coordinates(x, dim=2, t=t,
                                                             tau_w=tau_w)
                    tau_w = np.sqrt(3*x.var()/(y0[:, 1]**2).mean())
                print("tau_w set to", tau_w)
            if tau_w is None:
                p = dim
            else:
                p = 1
                while (t[2*p+1:] - t[:-(2*p+1)]).mean() < tau_w and p < N/4:
                    p += 1
                print("p set to", p)

        m = 2*p + 1
        N1 = N - m + 1

        # time differences:
        dt = np.zeros((N1, m))
        for i in range(N1):
            dt[i, :] = t[i:i+m] - t[i+p]

        # filter weights
        # = recursively computed values of orthogonal polynomials:
        r = np.zeros((N1, dim, m))
        for j in range(dim):
            r[:, j, :] = dt**j - (
                r[:, :j, :] * ((dt**j).reshape((N1, 1, m)) * r[:, :j, :]).sum(
                    axis=2).reshape((N1, j, 1))).sum(axis=1)
            r[:, j, :] /= np.sqrt((r[:, j, :]**2).sum(axis=1)).reshape((N1, 1))
        for j in range(dim):
            r[:, j, :] *= factorial(j) / \
                (r[:, j, :] * dt**j).sum(axis=1).reshape((-1, 1))

        # reconstructed state vectors = filtered time series values:
        y = np.zeros((N1, dim))
        for i in range(N1):
            y[i, :] = (r[i, :, :]*x[i:i+m].reshape((1, m))).sum(axis=1)

        return y

    @staticmethod
    def embed_time_series(time_series, dim, tau):
        """
        Return a time series' delay embedding.

        Returns a Numpy array containing a delay embedding of the time series
        using embedding dimension dim and time delay tau.

        :type time_series: 1D array
        :arg time_series: The scalar time series to be embedded.
        :arg int dim: The embedding dimension.
        :arg int tau: The embedding delay.
        :rtype: 2D array (time, dimension)
        :return: the embedded phase space trajectory.
        """
        #  Make sure that dim and tau are Python integers
        dim = int(dim)
        tau = int(tau)
        time_series = to_cy(time_series, FIELD)
        if time_series.ndim > 1:
            time_series = time_series.squeeze(axis=-1)
        assert time_series.ndim == 1
        n_time = time_series.shape[0]

        embedding = np.empty((n_time - (dim - 1) * tau, dim), dtype=FIELD)
        _embed_time_series(n_time, dim, tau, time_series, embedding)
        return embedding

    def permutation_entropy(self, normalize=True):
        """
        Returns the permutation entropy for an embedded time series.
        An embedding of 3 <= embedding dimension <= 7 is recommended for this
        method.

        Reference: [Bandt2002]_

        :rtype: double
        :return: Permutation entropy of the embedded time series
        """
        if self.dim is None or self.tau is None:
            raise ValueError("Permutation Entropy only works for "
                             "one-dimensional embedded time series!")

        # Calculate order from embedding
        temp = self.embedding.argsort(axis=1)
        ranks = temp.argsort(axis=1)

        # Calculate probability distribution
        P = np.unique(ranks, return_counts=True, axis=0)[1]
        P = P / ranks.shape[0]

        entropy = abs(-(P*np.log2(P)).sum())
        if normalize:
            entropy = entropy / np.log2(factorial(self.dim))

        return entropy

    def complexity_entropy(self):
        """
        Returns the complexity entropy for each dimension of the time series.

        Reference: [Ribeiro2011]_

        :rtype: double
        :return: Complexity entropy of the embedded time series
        """

        if self.dim is None or self.tau is None:
            raise ValueError("Complexity Entropy only works for "
                             "one-dimensional embedded time series!")

        # Calculate ranks from embedding
        temp = self.embedding.argsort(axis=1)
        ranks = temp.argsort(axis=1)

        # Calculate probability distribution
        P = np.unique(ranks, return_counts=True, axis=0)[1]
        P = P / ranks.shape[0]

        # Calculate factorial of embedding dimension
        d_fac = factorial(self.dim)

        # Calculate permutation entropy
        S_P = abs(-(P*np.log2(P)).sum())

        # Initialize unit probability distribution
        P_e = np.ones(d_fac) / d_fac

        # Calculate permutation entropy of unit probability distribution
        S_Pe = abs(-(P_e*np.log2(P_e)).sum())

        # Calculate combined probability distribution
        P_comb = np.zeros(d_fac)
        P_comb[:len(P)] = P
        P_comb = (P_comb + P_e) / 2

        # Calculate permutation entropy of combinded probability distribution
        S_PPe = abs(-(P_comb*np.log2(P_comb)).sum())

        # Calculation of maximum Q split into several steps for
        # readability
        Q_max = (d_fac + 1) / d_fac * np.log2(d_fac + 1)
        Q_max = Q_max - 2*np.log2(2*d_fac) + np.log2(d_fac)
        Q_max = - Q_max / 2

        Q = (S_PPe - S_P/2 - S_Pe/2) / Q_max

        complexity_entropy = Q * S_P / np.log2(d_fac)

        return complexity_entropy

    #
    #  Calculate recurrence plot
    #

    @Cached.method(name="the manhattan distance matrix")
    def manhattan_distance_matrix(self):
        """
        Return the manhattan distance matrix from an embedding of a time
        series.

        :rtype: 2D square array
        :return: the manhattan distance matrix.
        """
        (n_time, dim) = self.embedding.shape
        return _manhattan_distance_matrix_rp(n_time, dim, self.embedding)

    @Cached.method(name="the euclidean distance matrix")
    def euclidean_distance_matrix(self):
        """
        Return the euclidean distance matrix from an embedding of a time
        series.

        :rtype: 2D square array
        :return: the euclidean distance matrix.
        """
        (n_time, dim) = self.embedding.shape
        return _euclidean_distance_matrix_rp(n_time, dim, self.embedding)

    @Cached.method(name="the supremum distance matrix")
    def supremum_distance_matrix(self):
        """
        Return the supremum distance matrix from an embedding of a time series.

        :rtype: 2D square Numpy array
        :return: the supremum distance matrix.
        """
        (n_time, dim) = self.embedding.shape
        return _supremum_distance_matrix_rp(n_time, dim, self.embedding)

    def set_fixed_threshold(self, threshold):
        """
        Set the recurrence plot to a fixed threshold.

        Modifies / sets the class variables :attr:`R` and :attr:`N`
        accordingly.

        :arg number threshold: The recurrence threshold.
        """
        if self.silence_level <= 1:
            print("Calculating recurrence plot at fixed threshold...")

        distance = RecurrencePlot.distance_matrix(self, self.metric)
        n_time = distance.shape[0]
        recurrence = np.zeros((n_time, n_time), dtype="int8")
        recurrence[distance < threshold] = 1
        if self.missing_values:
            #  Write missing value lines and rows to recurrence matrix
            #  NaN flag is not supported by int8 data format -> use 0 here
            recurrence[self.missing_value_indices, :] = 0
            recurrence[:, self.missing_value_indices] = 0

        self.R = recurrence

    def set_fixed_threshold_std(self, threshold_std):
        """
        Set the recurrence plot to a fixed threshold in units of the standard
        deviation of the time series.

        Calculates the absolute threshold and calls
        :meth:`set_fixed_threshold`.

        :arg number threshold_std: The recurrence threshold in units of the
            standard deviation of the time series.
        """
        if self.silence_level <= 1:
            print("Calculating recurrence plot at fixed threshold in units of "
                  "time series STD...")

        threshold = threshold_std * self.time_series.std()
        RecurrencePlot.set_fixed_threshold(self, threshold)

    def set_fixed_recurrence_rate(self, recurrence_rate):
        """
        Set the recurrence plot to a fixed recurrence rate.

        Modifies / sets the class variables :attr:`R` and :attr:`N`
        accordingly.

        :arg number recurrence_rate: The recurrence rate.
        """
        if self.silence_level <= 1:
            print("Calculating recurrence plot at fixed recurrence rate...")

        distance = RecurrencePlot.distance_matrix(self, self.metric)
        n_time = distance.shape[0]
        threshold = self.threshold_from_recurrence_rate(distance,
                                                        recurrence_rate)
        recurrence = np.zeros((n_time, n_time), dtype="int8")
        recurrence[distance < threshold] = 1
        self.R = recurrence

    def set_fixed_local_recurrence_rate(self, local_recurrence_rate):
        """
        Set the recurrence plot to a fixed local recurrence rate.

        This results in a fixed number of recurrences for each state vector,
        i.e., all state vectors have the same number of recurrences.  Modifies
        / sets the class variables :attr:`R` and :attr:`N` accordingly.

        .. note::
           The resulting recurrence matrix :math:`R` is generally asymmetric!

        :arg number local_recurrence_rate: The local recurrence rate.
        """
        if self.silence_level <= 1:
            print("Calculating recurrence plot at fixed "
                  "local recurrence rate...")

        distance = RecurrencePlot.distance_matrix(self, self.metric)
        n_time = distance.shape[0]
        recurrence = np.zeros((n_time, n_time), dtype="int8")
        for i in range(n_time):
            #  Get threshold for state vector i to obtain fixed local
            #  recurrence rate
            local_threshold = self.threshold_from_recurrence_rate(
                distance[i, :], local_recurrence_rate)
            #  Thresholding the distance matrix for column i
            recurrence[i, distance[i, :] < local_threshold] = 1
        self.R = recurrence

    def set_adaptive_neighborhood_size(self, adaptive_neighborhood_size,
                                       order=None):
        """
        Construct recurrence plot using the :index:`adaptive neighborhood
        size <single: adaptive neighborhood size; recurrence plot>` algorithm
        introduced in [Xu2008]_.

        The exact algorithm was deduced from private correspondence with the
        authors, as the description given in the above mentioned is not correct
        or at least ambiguous.

        Modifies / sets the class variables :attr:`R` and :attr:`N`
        accordingly.

        :arg number adaptive_neighborhood_size: The number of adaptive nearest
            neighbors (recurrences) assigned to each state vector.
        :type order: 1D array of int32
        :arg order: The indices of state vectors in the order desired for
            processing by the algorithm. The standard order is :math:`1,...,N`.
        """
        if self.silence_level <= 1:
            print("Calculating recurrence plot using the "
                  "adaptive neighborhood size algorithm...")

        distance = RecurrencePlot.distance_matrix(self, self.metric)

        #  Get indices that would sort the distance matrix.
        #  sorted_neighbors[i,j] contains the index of the jth nearest neighbor
        #  of i. Sorting order is very important here!
        sorted_neighbors = to_cy(distance.argsort(axis=1), NODE)

        n_time = distance.shape[0]
        recurrence = np.zeros((n_time, n_time), dtype=LAG)

        #  Set processing order of state vectors
        if order is None:
            order = np.arange(n_time, dtype=NODE)
        else:
            order = to_cy(order, NODE)

        _set_adaptive_neighborhood_size(n_time, adaptive_neighborhood_size,
                                        sorted_neighbors, order, recurrence)
        self.R = recurrence

    @staticmethod
    def threshold_from_recurrence_rate(distance, recurrence_rate: float):
        """
        Return the threshold for recurrence plot construction given the
        recurrence rate.

        Be aware, that the returned threshold can only approximately give the
        desired recurrence rate. The accuracy depends on the distribution of
        values in the given distance matrix :math:`D`.

        :type distance: 2D square array.
        :arg distance: The phase space distance matrix :math:`D`.
        :arg number recurrence_rate: The desired recurrence rate.
        :return number: the recurrence threshold corresponding to the desired
            recurrence rate.
        """
        #  Flatten and sort distance matrix
        flat_distance = distance.flatten()
        flat_distance.sort()

        #  Get threshold
        assert 0 <= recurrence_rate <= 1
        N = len(flat_distance)
        threshold = flat_distance[int(recurrence_rate * (N - 1))]

        #  Clean up
        del flat_distance

        return threshold

    @staticmethod
    def threshold_from_recurrence_rate_fast(distance, recurrence_rate,
                                            rr_precision=0.001):
        """
        Return the threshold for recurrence plot construction given the
        recurrence rate.

        The threshold yielding a given recurrence_rate is approximated using a
        randomly selected rr_precision percent of the distance matrix' entries.
        Hence, the expected accuracy is lower than that achieved by using
        :index:`threshold_from_recurrence_rate`.

        :type distance: 2D square array.
        :arg distance: The phase space distance matrix :math:`D`.
        :arg number recurrence_rate: The desired recurrence rate.
        :arg number rr_precision: The desired precision of recurrence rate
            estimation.
        :return number: the recurrence threshold corresponding to the desired
            recurrence rate.
        """
        #  Get number of distances to be randomly chosen
        n_samples = int(rr_precision * distance.size)

        #  Get number of phase space points
        n_time = distance.shape[0]

        # vectorized version
        i = np.random.randint(n_time, size=n_samples)
        j = np.random.randint(n_time, size=n_samples)
        samples = distance[i, j]

        #  Sort and get threshold
        samples.sort()
        threshold = samples[int(recurrence_rate * n_samples)]
        return threshold

    @staticmethod
    def bootstrap_distance_matrix(embedding, metric, M):
        """
        Return bootstrap samples from distance matrix.

        :type embedding: 2D array (time, embedding dimension)
        :arg embedding: The phase space trajectory.
        :arg str metric: The metric for measuring distances in phase space
            ("manhattan", "euclidean", "supremum").
        :arg int M: Number of bootstrap samples
        :rtype: 1D array ("float32")
        :return: the bootstrap samples from distance matrix.
        """
        #  Prepare
        M = int(M)
        embedding = to_cy(embedding, DFIELD)
        distances = np.zeros(M, dtype=DFIELD)
        (n_time, dim) = embedding.shape

        if metric == "manhattan":
            _bootstrap_distance_matrix_manhattan(n_time, dim, embedding,
                                                 distances, M)

        elif metric == "euclidean":
            _bootstrap_distance_matrix_euclidean(n_time, dim, embedding,
                                                 distances, M)

        elif metric == "supremum":
            _bootstrap_distance_matrix_supremum(n_time, dim, embedding,
                                                distances, M)
        return distances

    #
    #  Recurrence quantification analysis (RQA)
    #

    def rqa_summary(self, l_min=2, v_min=2):
        """
        Return a selection of RQA measures.

        The selection consists of the recurrence rate :math:`RR`, the
        determinism :math:`DET`, the average diagonal line length :math:`L` and
        the laminarity :math:`LAM`.

        :arg int l_min: The minimum diagonal line length.
        :arg int v_min: The minimum vertical line length.
        :rtype: Python dictionary
        :return: a selection of RQA measures.
        """
        RR = self.recurrence_rate()
        DET = self.determinism(l_min)
        L = self.average_diaglength(l_min)
        LAM = self.laminarity(v_min)

        return {"RR": RR, "DET": DET, "L": L, "LAM": LAM}

    def recurrence_rate(self):
        """
        Return the :index:`recurrence rate` :math:`RR`.

        RR gives the percentage of black dots in the recurrence plot.

        :return number: the recurrence rate :math:`RR`.
        """
        N = self.N
        if not self.sparse_rqa:
            R = self.recurrence_matrix()
            RR = R.sum() / N ** 2
        elif self.metric == "supremum":
            RR = (self.vertline_dist() * np.arange(1, N + 1)).sum() / N ** 2
        else:
            raise NotImplementedError(
                "Sequential RQA is currently only available for "
                "fixed threshold and the supremum metric.")
        return RR

    def recurrence_probability(self, lag=0):
        """
        Return the recurrence probability. This is the probability, that
        the trajectory is recurrent after 'lag' time steps.

        Contributed by Jan H. Feldhoff.

        :return number: the recurrence probability
        """
        R = self.recurrence_matrix()
        N = self.N
        SUM = np.sum(np.diag(R, lag))

        return SUM / float(N-lag)

    #
    #  RQA measures based on black diagonal lines
    #

    @Cached.method(attrs=(
        "metric", "threshold", "missing_values", "sparse_rqa"))
    def diagline_dist(self):
        """
        Return the :index:`frequency distribution of diagonal line lengths
        <triple: frequency distribution; diagonal; line length>`
        :math:`P(l-1)`.

        Note that entry :math:`P(l-1)` contains the number of
        :index:`diagonal lines <pair: diagonal; lines>` of length :math:`l`.
        Thus, :math:`P(0)` counts lines of length :math:`1`,
        :math:`P(1)` counts lines of length :math:`2`, asf.
        The main diagonal is not counted,
        hence :math:`P(N)` will always be :math:`0`.

        .. note::
           Experimental handling of missing values. Diagonal lines
           touching lines and blocks of missing entries in the
           recurrence matrix are not counted.

        :rtype: 1D array (int32)
        :return: the frequency distribution of diagonal line lengths
            :math:`P(l-1)`.
        """
        #  Prepare
        n_time = self.N
        diagline = np.zeros(n_time, dtype=NODE)

        if not self.sparse_rqa:
            #  Get recurrence matrix
            recmat = self.recurrence_matrix()

            if self.missing_values:
                mv_indices = self.missing_value_indices
                _diagline_dist_missingvalues(
                    n_time, diagline, recmat, mv_indices)
            else:
                _diagline_dist(n_time, diagline, recmat)

        #  Calculations for sequential RQA
        elif self.metric == "supremum" and self.threshold is not None:
            #  Get embedding
            embedding = self.embedding
            #  Get time series dimension
            dim = embedding.shape[1]
            #  Get threshold
            eps = float(self.threshold)

            if self.missing_values:
                mv_indices = self.missing_value_indices
                _diagline_dist_sequential_missingvalues(
                    n_time, diagline, mv_indices, embedding, eps, dim)
            else:
                _diagline_dist_sequential(
                    n_time, diagline, embedding, eps, dim)

        else:
            raise NotImplementedError(
                "Sequential RQA is currently only available for "
                "fixed threshold and the supremum metric.")

        #  Function just runs over the upper triangular matrix
        return 2 * diagline

    @staticmethod
    def rejection_sampling(dist, M):
        """
        Rejection sampling of discrete frequency distribution.

        Use simple rejection sampling algorithm for computing a resampled
        version of a given frequency distribution with discrete support.

        :type dist: 1D array (integer)
        :arg dist: discrete frequency distribution
        :arg int M: number of resamplings
        :rtype: 1D array (integer)
        :return: the resampled frequency distribution.
        """
        #  Get number of support points
        N = len(dist)

        #  Prepare
        resampled_dist = np.zeros(N, dtype=NODE)

        #  Prescribed distribution
        dist = to_cy(dist, DFIELD)
        #  Normalize distribution
        dist /= dist.sum()

        _rejection_sampling(dist, resampled_dist, N, M)
        return resampled_dist

    def resample_diagline_dist(self, M):
        """
        Return resampled frequency distribution of diagonal lines.

        The resampled frequency distribution can be used for obtaining
        confidence bounds on diagonal line based RQA measures. This is
        described in detail in [Schinkel2009]_.

        Concerning the choice of the number of resamplings, Schinkel et al.
        write: "The number of resamplings is not generally agreed upon but
        common guidelines suggest values between 800 and 1500."

        :arg int M: number of resamplings
        :rtype: 1D array (integer)
        :return: the resampled frequency distribution of diagonal lines.
        """
        #  Get original distribution of diagonal lines
        diagline = self.diagline_dist()

        #  Get maximal diagonal line length
        L_max = self.max_diaglength()

        #  Get resampled distribution
        if L_max == 0:
            resampled_dist = diagline
        else:
            resampled_dist = np.zeros(len(diagline), dtype=NODE)
            resampled_dist[:L_max] = RecurrencePlot.\
                rejection_sampling(diagline[:L_max], M)

        return resampled_dist

    def max_diaglength(self):
        """
        Return diagonal line-based RQA measure :index:`maximum diagonal line
        length <triple: maximum; diagonal; line length>` :math:`L_max`.

        :math:`L_max` is defined as the maximal length of a diagonal line in
        the recurrence matrix.

        :return number: the maximal diagonal line length :math:`L_max`.
        """
        return 1 + np.nonzero(self.diagline_dist())[0].max(initial=-1)

    def determinism(self, l_min=2, resampled_dist=None):
        """
        Return diagonal line-based RQA measure :index:`determinism <pair: RQA;
        determinism>` :math:`DET`.

        :math:`DET` is defined as the ratio of recurrence points that form
        diagonal structures (of at least length :math:`l_min`) to all
        recurrence points.

        :arg number l_min: The minimum diagonal line length.
        :type resampled_dist: 1D array (integer)
        :arg resampled_dist: resampled frequency distribution of diagonal lines
        :return number: the determinism :math:`DET`.
        """
        #  Use resampled distribution of diagonal lines if provided
        diagline = (self.diagline_dist() if resampled_dist is None
                    else resampled_dist)
        n_time = self.N

        #  Number of recurrence points that form diagonal structures (of at
        #  least length l_min)
        partial_sum = np.arange(l_min, n_time+1) @ diagline[l_min-1:]

        #  Number of all recurrence points that form diagonal lines (except
        #  the main diagonal)
        full_sum = np.arange(1, n_time+1) @ diagline

        return partial_sum / float(full_sum + self._epsilon)

    def average_diaglength(self, l_min=2, resampled_dist=None):
        """
        Return diagonal line-based RQA measure :index:`average diagonal line
        length <triple: average; diagonal; line length>` :math:`L`.

        :math:`L` is defined as the average length of diagonal lines (of at
        least length :math:`l_min`).

        :arg number l_min: The minimum diagonal line length.
        :type resampled_dist: 1D array (integer)
        :arg resampled_dist: resampled frequency distribution of diagonal lines
        :return number: the average diagonal line length :math:`L`.
        """
        #  Use resampled distribution of diagonal lines if provided
        diagline = (self.diagline_dist() if resampled_dist is None
                    else resampled_dist)
        n_time = self.N

        #  Number of recurrence points that form diagonal structures (of at
        #  least length l_min)
        partial_sum = np.arange(l_min, n_time+1) @ diagline[l_min-1:]

        #  Total number of diagonal lines of at least length l_min
        number_diagline = diagline[l_min-1:].sum()

        return partial_sum / float(number_diagline + self._epsilon)

    def diag_entropy(self, l_min=2, resampled_dist=None):
        """
        Return diagonal line-based RQA measure :index:`diagonal line entropy
        <pair: diagonal; line entropy>` :math:`ENTR`.

        :math:`ENTR` is defined as the entropy of the probability to find a
        diagonal line of exactly length l in the RP - reflects the complexity
        of the RP with respect to diagonal lines.

        :arg number l_min: The minimal diagonal line length.
        :type resampled_dist: 1D array (integer)
        :arg resampled_dist: resampled frequency distribution of diagonal lines
        :return number: the diagonal line-based entropy :math:`ENTR`.
        """
        #  Use resampled distribution of diagonal lines if provided
        diagline = (self.diagline_dist() if resampled_dist is None
                    else resampled_dist)

        #  Creates a reduced array of the values (not 0) of the diagonal line
        #  length (langer than l_min)
        diagline = diagline[l_min-1:]
        diagline = np.extract(diagline != 0, diagline)

        #  Normalized array of the number of all diagonal lines = probability
        #  of diagonal line length
        diagnorm = diagline / float(diagline.sum() + self._epsilon)

        return -(diagnorm * np.log(diagnorm)).sum()

    #
    #  RQA measures based on black vertical lines
    #

    @Cached.method(attrs=(
        "metric", "threshold", "missing_values", "sparse_rqa"))
    def vertline_dist(self):
        """
        Return the :index:`frequency distribution of vertical line lengths
        <triple: frequency distribution; vertical; line length>`
        :math:`P(v-1)`.

        Note that entry :math:`P(v-1)` contains the number of
        :index:`vertical lines <pair: vertical; lines>` of length :math:`v`.
        Thus, :math:`P(0)` counts lines of length :math:`1`,
        :math:`P(1)` counts lines of length :math:`2`, asf.

        :rtype: 1D array (int32)
        :return: the frequency distribution of vertical line lengths
            :math:`P(v-1)`.
        """
        #  Prepare
        n_time = self.N
        vertline = np.zeros(n_time, dtype=NODE)

        if not self.sparse_rqa:
            #  Get recurrence matrix
            recmat = self.recurrence_matrix()

            if self.missing_values:
                mv_indices = self.missing_value_indices
                _vertline_dist_missingvalues(
                    n_time, vertline, recmat, mv_indices)
            else:
                _vertline_dist(n_time, vertline, recmat)

        #  Calculations for sequential RQA
        elif self.metric == "supremum" and self.threshold is not None:
            #  Get embedding
            embedding = self.embedding
            #  Get time series dimension
            dim = embedding.shape[1]
            #  Get threshold
            eps = float(self.threshold)

            if self.missing_values:
                mv_indices = self.missing_value_indices
                _vertline_dist_sequential_missingvalues(
                    n_time, vertline, mv_indices, embedding, eps, dim)

            else:
                _vertline_dist_sequential(
                    n_time, vertline, embedding, eps, dim)

        else:
            raise NotImplementedError(
                "Sequential RQA is currently only available for "
                "fixed threshold and the supremum metric.")

        #  Function covers the whole recurrence matrix
        return vertline

    def resample_vertline_dist(self, M):
        """
        Return resampled frequency distribution of vertical lines.

        The resampled frequency distribution can be used for obtaining
        confidence bounds on vertical line based RQA measures. This is
        described in detail in [Schinkel2009]_.

        Concerning the choice of the number of resamplings, Schinkel et al.
        write: "The number of resamplings is not generally agreed upon but
        common guidelines suggest values between 800 and 1500."

        :arg int M: number of resamplings
        :rtype: 1D array (integer)
        :return: the resampled frequency distribution of vertical lines.
        """
        #  Get original distribution of vertical lines
        vertline = self.vertline_dist()

        #  Get maximal vertical line length
        L_max = self.max_vertlength()

        #  Get resampled distribution
        if L_max == 0:
            resampled_dist = vertline
        else:
            resampled_dist = np.zeros(len(vertline), dtype=NODE)
            resampled_dist[:L_max] = RecurrencePlot.\
                rejection_sampling(vertline[:L_max], M)

        return resampled_dist

    def max_vertlength(self):
        """
        Return vertical line-based RQA measure :index:`maximal vertical line
        length <triple: maximum; vertical; line length>` :math:`V_max`.

        :math:`V_max` is defined as the maximal length of a vertical line of
        the recurrence matrix.

        :return number: the maximal vertical line length :math:`V_max`.
        """
        return 1 + np.nonzero(self.vertline_dist())[0].max(initial=-1)

    def laminarity(self, v_min=2, resampled_dist=None):
        """
        Return vertical line-based RQA measure :index:`laminarity` :math:`LAM`.

        :math:`LAM` is defined as the ratio of recurrence points that form
        vertical structures (of at least length :math:`v_min`) to all
        recurrence points.

        :arg number v_min: The minimal vertical line length.
        :type resampled_dist: 1D array (integer)
        :arg resampled_dist: resampled frequency distribution of vertical lines
        :return number: the laminarity :math:`LAM`.
        """
        #  Use resampled distribution of vertical lines if provided
        vertline = (self.vertline_dist() if resampled_dist is None
                    else resampled_dist)
        n_time = self.N

        #  Number of recurrence points that form vertical structures (of at
        #  least length v_min)
        partial_sum = np.arange(v_min, n_time+1) @ vertline[v_min-1:]

        #  Number of all recurrence points that form vertical lines
        full_sum = np.arange(1, n_time+1) @ vertline

        return partial_sum / float(full_sum + self._epsilon)

    def average_vertlength(self, v_min=2, resampled_dist=None):
        """
        Return vertical line-based RQA measure :index:`average vertical line
        length <triple: average; vertical; line length>` :math:`TT`.

        :math:`TT` is defined as the average vertical line length (of at least
        length :math:`v_min`) and is also called :index:`trapping time`
        :math:`TT`.

        :arg number v_min: The minimal vertical line length.
        :type resampled_dist: 1D array (integer)
        :arg resampled_dist: resampled frequency distribution of vertical lines
        :return number: the trapping time :math:`TT`.
        """
        #  Use resampled distribution of vertical lines if provided
        vertline = (self.vertline_dist() if resampled_dist is None
                    else resampled_dist)
        n_time = self.N

        #  Number of recurrence points that form vertical structures (of at
        #  least length v_min)
        partial_sum = np.arange(v_min, n_time+1) @ vertline[v_min-1:]

        #  Total number of vertical lines of at least length v_min
        number_vertline = vertline[v_min-1:].sum()

        return partial_sum / (float(number_vertline) + self._epsilon)

    def trapping_time(self, v_min=2, resampled_dist=None):
        """
        Alias for :meth:`average_vertlength` (see description there).
        """
        return self.average_vertlength(v_min, resampled_dist)

    def vert_entropy(self, v_min=2, resampled_dist=None):
        """
        Return vertical line-based RQA measure :index:`vertical line entropy
        <pair: vertical; line entropy>`.

        It is defined as the entropy of the probability to find a vertical line
        of exactly length l in the RP - reflects the complexity of the RP with
        respect to vertical lines.

        :arg int v_min: The minimal vertical line length.
        :type resampled_dist: 1D array (integer)
        :arg resampled_dist: resampled frequency distribution of vertical lines
        :return number: the vertical line-based entropy.
        """
        #  Use resampled distribution of vertical lines if provided
        if resampled_dist is None:
            vertline = self.vertline_dist()
        else:
            vertline = resampled_dist

        #  Creates a reduced array of the values (not 0) of the vertical line
        #  length (langer than v_min)
        vertline = vertline[v_min-1:]
        vertline = np.extract(vertline != 0, vertline)

        #  Normalized array of the number of all vertical lines = probability
        #  of vertical line length
        vertline_normed = vertline / float(vertline.sum() + self._epsilon)

        return -(vertline_normed * np.log(vertline_normed)).sum()

    #
    #  RQA measures based on white vertical lines
    #

    def white_vertline_dist(self):
        """
        Return the :index:`frequency distribution of white vertical line
        lengths <triple: frequency distribution; white vertical; line length>`
        :math:`P(w-1)`.

        Note that entry :math:`P(w-1)` contains the number of
        :index:`white vertical lines <pair: white vertical; lines>` of length
        :math:`w`. Thus, :math:`P(0)` counts lines of length :math:`1`,
        :math:`P(1)` counts lines of length :math:`2`, asf.

        The length of a white vertical line in a recurrence plot corresponds to
        the time the system takes to return close to an earlier state.

        :rtype: 1D array (int32)
        :return: the frequency distribution of white vertical line lengths
            :math:`P(w-1)`.
        """
        R = self.recurrence_matrix()
        n_time = self.N
        white_vertline = np.zeros(n_time, dtype=NODE)
        _white_vertline_dist(n_time, white_vertline, R)

        #  Function covers the whole recurrence matrix
        return white_vertline

    def max_white_vertlength(self):
        """
        Return white vertical line-based RQA measure :index:`maximal white
        vertical line length <triple: maximum; white vertical; line length>`.

        It is defined as the maximal length of a white vertical line of
        the recurrence matrix and corresponds to the maximum recurrence time
        occuring in the time series.

        :return number: the maximal white vertical line length.
        """
        return 1 + np.nonzero(self.white_vertline_dist())[0].max(initial=-1)

    def average_white_vertlength(self, w_min=1):
        """
        Return white vertical line-based RQA measure :index:`average white
        vertical line length <triple: average; white vertical; line length>`.

        It is defined as the average white vertical line length (of at least
        length :math:`w_min`) and is also called :index:`mean recurrence time`.

        Reference: [Ngamga2007]_.

        :arg number w_min: The minimal white vertical line length.
        :return number: the mean recurrence time.
        """
        white_vertline = self.white_vertline_dist()
        n_time = self.N

        #  Number of recurrence points that form white vertical structures
        #  (of at least length w_min)
        partial_sum = np.arange(w_min, n_time+1) @ white_vertline[w_min-1:]

        #  Total number of white vertical lines of at least length v_min
        number_white_vertline = white_vertline[w_min-1:].sum()

        return partial_sum / float(number_white_vertline + self._epsilon)

    def mean_recurrence_time(self, w_min=1):
        """
        Alias for :meth:`average_white_vertlength` (see description there).
        """
        return self.average_white_vertlength(w_min)

    def white_vert_entropy(self, w_min=1):
        """
        Return white vertical line-based RQA measure :index:`white vertical
        line entropy <pair: white vertical; line entropy>`.

        It is defined as the entropy of the probability to find a white
        vertical line of exactly length l in the RP - reflects the complexity
        of the RP with respect to white vertical lines (recurrence times).

        :arg int w_min: Minimal white vertical line length (recurrence time).
        :return number: the white vertical line-based entropy.
        """
        #  Creates a reduced array of the values (not 0) of the vertical line
        #  length (langer than v_min)
        white_vertline = self.white_vertline_dist()
        white_vertline = white_vertline[w_min-1:]
        white_vertline = np.extract(white_vertline != 0, white_vertline)

        #  Normalized array of the number of all vertical lines = probability
        #  of vertical line length
        white_vertline_normed = white_vertline / float(
            white_vertline.sum() + self._epsilon)

        return -(white_vertline_normed * np.log(white_vertline_normed)).sum()

    #
    #  Methods for recurrence-based surrogates
    #

    def twins(self, min_dist=7):
        """
        Return list of the :index:`twins <pair: twins; recurrence plot>` of
        each state vector based on the recurrence matrix.

        Two state vectors are said to be twins if they share the same
        recurrences, i.e., if the corresponding rows or columns in the
        recurrence plot are identical.

        References: [Thiel2006]_, [Marwan2007]_.

        :arg number min_dist: The minimum temporal distance for twins.
        :return [[number]]: the list of twins for each state vector in the time
            series.
        """
        if self.silence_level <= 1:
            print("Finding twins based on recurrence matrix...")

        #  Initialize
        twins = []
        N = self.N

        #  Get current recurrence matrix
        R = self.recurrence_matrix()
        #  Get number of neighbors for each state vector
        nR = R.sum(axis=0)

        _twins_r(min_dist, N, R, nR, twins)
        return twins

    def twin_surrogates(self, n_surrogates=1, min_dist=7):
        """
        Generate surrogates based on the current (embedded) time series
        :attr:`embedding` using the :index:`twin surrogate` method.

        The twins surrogates have the same dimensionality as the (embedded)
        trajectory used for constructing the recurrence plot. If scalar
        surrogate time series are desired, any component of the twin surrogate
        trajectory may be isolated.

        Twin surrogates share linear and nonlinear properties with the original
        time series, since they correspond to realizations of trajectories of
        the same dynamical systems with different initial conditions.

        References: [Thiel2006]_ [*], [Marwan2007]_.

        :arg number min_dist: The minimum temporal distance for twins.
        :arg int n_surrogates: The number of twin surrogate trajectories to be
            returned.
        :rtype: 3D array (surrogate number, time, dimension)
        :return: the twin surrogate trajectories.
        """
        #  The algorithm proceeds in two steps:
        #  1. Use the algorithm proposed in [*] to find twins
        #  2. Reconstruct one-dimensional twin surrogate time series
        if self.silence_level <= 1:
            print("Generating twin surrogates...")

        #  Collect
        N = self.N
        embedding = self.embedding
        dim = embedding.shape[1]
        twins = self.twins(min_dist)

        return _twin_surrogates_r(n_surrogates, N, dim, twins, embedding)
